# RCAIDE/Framework/External_Interfaces/OpenVSP/get_vsp_measurements.py
# 
# Created:  --- 2016, T. MacDonald
# Modified: Aug 2017, T. MacDonald
#           Mar 2018, T. MacDonald
#           Jan 2020, T. MacDonald
#           Feb 2021, T. MacDonald

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
try:
    import vsp as vsp
except ImportError:
    try:
        import openvsp as vsp
    except ImportError:
        # This allows RCAIDE to build without OpenVSP
        pass
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Get VSP Measurements
# ---------------------------------------------------------------------------------------------------------------------- 
def get_vsp_measurements(filename = 'Unnamed_CompGeom.csv', measurement_type = 'wetted_area'):
    """This calls OpenVSP to compute the wetted areas or volumes of a previously written vehicle.
    
    Assumptions:
    Vehicle must be open in OpenVSP (via recently used vsp_write)
    All components have different tags. Repeated tags are added together under the
    assumption that this represents multiple engines or similar. Values computed from
    repeated tags in this way may need to be divided by the number of entities later 
    before assignment. This is because some analyses may multiply an assigned area
    by number of engines, for example.

    Source:
    N/A

    Inputs:
    None

    Outputs:
    measurement   [m^2 or m^3] - Dictionary with values for each component, 
      with component tags as the keys.

    Properties Used:
    N/A
    """        
    
    if measurement_type == 'wetted_area':
        output_ind = 2
    elif measurement_type == 'wetted_volume':
        output_ind = 4
    else:
        raise NotImplementedError
    
    half_mesh = False # Note that this does not affect the Gmsh/SU2 meshing process
    # it only affects how much area of a component is included in the output
    try:
        file_type = vsp.COMP_GEOM_CSV_TYPE
    except NameError:
        print('VSP import failed')
        return -1

    vsp.SetComputationFileName(file_type, filename)
    vsp.ComputeCompGeom(vsp.SET_ALL, half_mesh, file_type)
    
    f = open(filename)
    
    measurements = dict()
    
    # Extract wetted areas for each component
    for ii, line in enumerate(f):
        if ii == 0:
            pass
        elif line == '\n':
            break
        else:
            vals = line.split(',')
            item_tag = vals[0][:]
            item_w_area = float(vals[output_ind])
            if item_tag in measurements:
                item_w_area = measurements[item_tag] + item_w_area
            measurements[item_tag] = item_w_area
            
    f.close()
    
    return measurements


# RCAIDE/Library/Methods/Propulsors/Converters/Engine/__init__.py
# (c) Copyright 2023 Aerospace Research Community LLC

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .append_engine_conditions           import append_engine_conditions         
from .compute_throttle_from_power        import compute_throttle_from_power
from .compute_power_from_throttle        import compute_power_from_throttle
from .compute_fuel_flow_rate_from_torque import compute_fuel_flow_rate_from_torque


# RCAIDE/Methods/Noise/Common/generate_microphone_locations.py
# 
# 
# Created:  Oct 2023, A. Molloy  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# package imports  
import numpy as np 

# ---------------------------------------------------------------------------------------------------------------------- 
#  generate_zero_elevation_microphone_locations
# ---------------------------------------------------------------------------------------------------------------------- 
def generate_zero_elevation_microphone_locations(settings):
    """This computes the absolute microphone/observer locations on level ground. 
            
    Assumptions:
        None

    Source:
        N/A  

    Inputs:   
    settings.
        min_x - minimum x coordinate of noise evaluation plane [meters]
        max_x - maximum x coordinate of noise evaluation plane [meters]
        min_y - minimum y coordinate of noise evaluation plane [meters]
        max_x - maximim y coordinate of noise evaluation plane [meters]
        N_x   - number of microphones on x-axis 
        N_y   - number of microphones on y-axis 
    
    Outputs: 
        gm_mic_locations   - cartesian coordiates of all microphones defined  [meters] 
    
    Properties Used:
        N/A       
    """       

    N_x                   = settings.microphone_x_resolution 
    N_y                   = settings.microphone_y_resolution
    num_gm                = N_x*N_y
    gm_mic_locations      = np.zeros((num_gm,3))     
    min_x                 = settings.microphone_min_x         
    max_x                 = settings.microphone_max_x         
    min_y                 = settings.microphone_min_y         
    max_y                 = settings.microphone_max_y   
    x_coords_0            = np.repeat(np.linspace(min_x,max_x,N_x)[:,np.newaxis],N_y, axis = 1)
    y_coords_0            = np.repeat(np.linspace(min_y,max_y,N_y)[:,np.newaxis],N_x, axis = 1).T
    z_coords_0            = np.zeros_like(x_coords_0) 
    gm_mic_locations[:,0] = x_coords_0.reshape(num_gm)
    gm_mic_locations[:,1] = y_coords_0.reshape(num_gm)
    gm_mic_locations[:,2] = z_coords_0.reshape(num_gm) 
     
    return gm_mic_locations 


 
from RCAIDE.Framework.Core import Data
from RCAIDE.Framework.Analyses import Analysis


# ----------------------------------------------------------------------
#  Analysis
# ---------------------------------------------------------------------- 
class Stability(Analysis):
    """This is the base class for stability analyses. It contains functions
    that are built into the default class.
    
    Assumptions:
    None
    
    Source:
    N/A
    """
     
    def __defaults__(self):
        """This sets the default values and methods for the analysis.         

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """   
        self.tag      = 'stability'
        self.vehicle  = Data()
        self.settings = Data()
 
        return





# RCAIDE/Methods/Energy/Sources/Battery/Common/find_total_mass_gain.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ---------------------------------------------------------------------------------------------------------------------- -Common
def find_total_mass_gain(battery):
    """finds the total mass of air that the battery 
    accumulates when discharged fully
    
    Assumptions:
    Earth Atmospheric composition
    
    Inputs:
    battery.maximum_energy [J]
    battery.
      mass_gain_factor [kg/W]
      
    Outputs:
      mdot             [kg]
    """ 
    
    mgain=battery.maximum_energy*battery.mass_gain_factor
    
    return mgain


# RCAIDE/Library/Methods/Propulsors/Converters/Engine/compute_throttle_from_power.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
 # RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core                                         import Units

# package imports
import numpy as np 

# ---------------------------------------------------------------------------------------------------------------------- 
#  calculate_throttle_from_power
# ----------------------------------------------------------------------------------------------------------------------    
def compute_throttle_from_power(engine, engine_conditions,conditions):
    """ The internal combustion engine output power and specific power consumption.
    The following perperties are computed:
    engine 
       .outputs.power_specific_fuel_consumption (numpy.ndarray): Power (brake) SFC [lbf/(HP · h) ]
       .outputs.fuel_flow_rate                  (numpy.ndarray): Fuel flow rate    [kg/s]
       .outputs.throttle                        (numpy.ndarray): throttle setting  [unitless]    
    
    Source:
        Gagg and Ferrar model (ref: S. Gudmundsson, 2014 - eq. 7-16)
    
    Assumtions:
        None 
    
    Args:
        engine
          .sea_level_power                        (float): sea-level power      [W]
          .flat_rate_altitude                     (float): flat rate altitude   [m]
          .power_specific_fuel_consumption        (float): PSFC                 [RPM]
          .throttle setting               (numpy.ndarray): throttle             [unitless]
          .inputs.speed                   (numpy.ndarray): angular velocity     [rad/s]
        freestream conditions 
          .altitude                               (float): altitde              [m]
          .delta_isa                              (float): altitude offset      [m]
          
    Returns:
        None 
    """

    # Unpack atmospheric conditions 
    delta_isa  = conditions.freestream.delta_ISA
    altitude   = conditions.freestream.altitude
    
    # Unpack engine operating conditions 
    PSLS     = engine.sea_level_power
    h_flat   = engine.flat_rate_altitude
    P        = engine_conditions.power*1.0
    PSFC     = engine.power_specific_fuel_consumption
    
    altitude_virtual = altitude - h_flat        
    altitude_virtual[altitude_virtual<0.] = 0.   
    atmo             = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_values_0    = atmo.compute_values(0,0) 
    rho0             = atmo_values_0.density[0,0] 
    atmo_values      = atmo.compute_values(altitude_virtual,delta_isa) 
    rho              = atmo_values.density 

    #Compute density ratio 
    sigma        = rho / rho0 
    Pavailable   = PSLS * (sigma - 0.117) / 0.883        
    Pavailable[h_flat > altitude] = PSLS 
 
    # Compute throttle 
    throttle = P/Pavailable 
    P[P<0.] = 0. 

    # Compute fuel flow rate
    SFC             = PSFC* Units['lb/hp/hr']
    a               = np.zeros_like(altitude)
    fuel_flow_rate  = np.fmax(P*SFC,a)
    
    # Store outputs 
    engine_conditions.power_specific_fuel_consumption = PSFC
    engine_conditions.fuel_flow_rate                  = fuel_flow_rate
    engine_conditions.throttle                        = throttle

    return



# RCAIDE/Library/Missions/Segments/Cruise/Constant_Throttle_Constant_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# Package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def unpack_unknowns(segment):
    
    # unpack unknowns
    unknowns   = segment.state.unknowns 
    accel_x    = unknowns.acceleration 
    time       = unknowns.elapsed_time
     
    # rescale time
    t_initial  = segment.state.conditions.frames.inertial.time[0,0]
    t_final    = t_initial + time  
    t_nondim   = segment.state.numerics.dimensionless.control_points
    time       = t_nondim * (t_final-t_initial) + t_initial     

    # build acceleration
    N          = segment.state.numerics.number_of_control_points
    a          = np.zeros((N, 3))
    a[:, 0]    = accel_x[:,0]
    
    # apply unknowns
    conditions = segment.state.conditions 
    conditions.frames.inertial.acceleration_vector  = a
    conditions.frames.inertial.time[:,0]            = time[:,0]
    
    return 

# ----------------------------------------------------------------------
#  Integrate Velocity
# ---------------------------------------------------------------------- 

def integrate_velocity(segment):
    
    # unpack 
    conditions = segment.state.conditions
    v0         = segment.air_speed_start 
    beta       = segment.sideslip_angle
    I          = segment.state.numerics.time.integrate
    a          = conditions.frames.inertial.acceleration_vector
    
    # compute x-velocity
    velocity_xy = v0 + np.dot(I, a)[:,0]   
    v_x         = np.cos(beta)*velocity_xy
    v_y         = np.sin(beta)*velocity_xy

    # pack velocity
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    
    return

# ----------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------    

def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constant throttle and constant altitude, allows for acceleration

    Source:
    N/A

    Inputs:
    segment.altitude                             [meters]
    segment.air_speed_start                      [meters/second]
    segment.air_speed_end                        [meters/second] 
    segment.state.numerics.number_of_control_points [int]

    Outputs:
    state.conditions.energy.throttle        [unitless]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]

    Properties Used:
    N/A
    """    
    # unpack inputs
    alt      = segment.altitude 
    v0       = segment.air_speed_start
    vf       = segment.air_speed_end    
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]   

    if v0  is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        v0 = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # avoid having zero velocity since aero and propulsion models need non-zero Reynolds number
    if v0 == 0.0: v0 = 0.01
    if vf == 0.0: vf = 0.01
    
    # intial and final speed cannot be the same
    if v0 == vf:
        vf = vf + 0.01
        
    # repack
    segment.air_speed_start = v0
    segment.air_speed_end   = vf
    
    # pack conditions   
    segment.state.conditions.freestream.altitude[:,0] = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down    
    
# ----------------------------------------------------------------------
#  Solve Residuals
# ----------------------------------------------------------------------    

def solve_velocity(segment):
    """ Calculates the additional velocity residual
    
        Assumptions:
        The vehicle accelerates, residual on forces and to get it to the final speed
        
        Inputs:
        segment.air_speed_end                  [meters/second]
        segment.state.conditions: 
            frames.inertial.velocity_vector    [meters/second] 
        segment.state.numerics.time.differentiate
            
        Outputs:
        segment.state.residuals:
            forces               [meters/second^2]
            final_velocity_error [meters/second] 

        Properties Used:
        N/A
                                
    """    

    # unpack inputs
    conditions = segment.state.conditions 
    vf         = segment.air_speed_end
    v          = conditions.frames.inertial.velocity_vector 
    
    segment.state.residuals.final_velocity_error = (np.linalg.norm(v[-1,:])- vf)

    return


# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_drag_components.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_drag_components(results,
                         save_figure=False,
                         show_legend= True,
                         save_filename="Drag_Components",
                         file_type=".png",
                        width = 11, height = 7):
    """This plots the drag components of the aircraft
    
    Assumptions:
    None
    
    Source:
    None
    
    Inputs:
    results.segments.condtions.aerodynamics.coefficients.drag
          parasite.total
          induced.total
          compressible.total
          miscellaneous.total
          
    Outputs:
    Plots
    
    Properties Used:
    N/A
    """ 
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     
     
    fig   = plt.figure(save_filename)
    axis_1 = plt.subplot(1,1,1)
    fig.set_size_inches(12,height)
    
    for i in range(len(results.segments)): 
        time   = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min 
        drag   = results.segments[i].conditions.aerodynamics.coefficients.drag 
        cdp    = drag.parasite.total[:,0]
        cdi    = drag.induced.total[:,0]
        cdc    = drag.compressible.total[:,0]
        cdm    = drag.miscellaneous.total[:,0] 
        cd     = drag.total[:,0]  
        
        if i ==  0:
            axis_1.plot(time, cdp, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = r'$C_{Dp}$') 
            axis_1.plot(time,cdi, color = line_colors[i], marker = ps.markers[1], linewidth = ps.line_width,  label = r'$C_{Di}$')  
            axis_1.plot(time, cdc, color = line_colors[i], marker = ps.markers[2], linewidth = ps.line_width,  label =r'$C_{Dc}$')  
            axis_1.plot(time, cdm, color = line_colors[i], marker = ps.markers[3], linewidth = ps.line_width,  label =r'$C_{Dm}$')  
            axis_1.plot(time, cd, color = line_colors[i], marker = ps.markers[5], linewidth = ps.line_width,  label =r'$C_D$')
        else:
            axis_1.plot(time, cdp, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
            axis_1.plot(time,cdi, color = line_colors[i], marker = ps.markers[1], linewidth = ps.line_width)
            axis_1.plot(time, cdc, color = line_colors[i], marker = ps.markers[2], linewidth = ps.line_width)
            axis_1.plot(time, cdm, color = line_colors[i], marker = ps.markers[3], linewidth = ps.line_width) 
            axis_1.plot(time, cd, color = line_colors[i], marker = ps.markers[5], linewidth = ps.line_width)
    
        set_axes(axis_1)            
        axis_1.set_xlabel('Time (mins)')
        axis_1.set_ylabel('Drag Compoments') 
        
    
    if show_legend:                    
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})    
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'Drag Components'      
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return fig 


# RCAIDE/Library/Components/Energy/Sources/Solar_Panels/Solar_Panel.py
# 
# 
# Created:  Oct 2024, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
from RCAIDE.Library.Components    import Component    

# ----------------------------------------------------------------------------------------------------------------------
#  Solar_Panel Class
# ----------------------------------------------------------------------------------------------------------------------
class Solar_Panel(Component):
    """
    Class for modeling photovoltaic solar panels in aircraft applications
    
    Attributes
    ----------
    area : float
        Solar panel surface area [m^2] (default: 0.0)
        
    efficiency : float
        Solar energy conversion efficiency [-] (default: 0.0)

    Notes
    -----
    The solar panel model provides basic functionality for converting solar
    radiation into electrical power based on panel area and efficiency.

    See Also
    --------
    RCAIDE.Library.Components.Energy.Modulators.Solar_Logic
        Solar power management and control
    """       
    def __defaults__(self):
        """
        Sets default values for solar panel attributes
        """        
        self.area       = 0.0
        self.efficiency = 0.0
    #def solar_radiation(self,conditions):  
        #"""Computes the adjusted solar flux

        #Assumptions:
        #Solar intensity =1305 W/m^2
        #Includes a diffuse component of 0% of the direct component
        #Altitudes are not excessive 

        #Source:
        #N/A

        #Inputs:
        #conditions.frames.
          #planet.start_time        [s]
          #planet.latitude          [degrees]
          #planet.longitude         [degrees]
          #body.inertial_rotations  [radians]
          #inertial.time            [s]
        #conditions.freestream.
          #altitude                 [m]

        #Outputs:
        #self.outputs.flux          [W/m^2]
        #flux                       [W/m^2]

        #Properties Used:
        #N/A
        #"""            
        
        ## Unpack
        #timedate  = conditions.frames.planet.start_time
        #latitude  = conditions.frames.planet.latitude
        #longitude = conditions.frames.planet.longitude
        #phip      = conditions.frames.body.inertial_rotations[:,0,None]
        #thetap    = conditions.frames.body.inertial_rotations[:,1,None]
        #psip      = conditions.frames.body.inertial_rotations[:,2,None]
        #altitude  = conditions.freestream.altitude
        #times     = conditions.frames.inertial.time
        
        ## Figure out the date and time
        #day       = timedate.tm_yday + np.floor_divide(times, 24.*60.*60.)
        #TUTC      = timedate.tm_sec + 60.*timedate.tm_min+ 60.*60.*timedate.tm_hour + np.mod(times,24.*60.*60.)
        
        ## Gamma is defined to be due south, so
        #gamma = psip - np.pi
        
        ## Solar intensity external to the Earths atmosphere
        #Io = 1367.0
        
        ## Indirect component adjustment
        #Ind = 1.0
        
        ## B
        #B = (360./365.0)*(day-81.)*np.pi/180.0
        
        ## Equation of Time
        #EoT = 9.87*np.sin(2*B)-7.53*np.cos(B)-1.5*np.sin(B)
        
        ## Time Correction factor
        #TC = 4*longitude+EoT
        
        ## Local Solar Time
        #LST = TUTC/3600.0+TC/60.0
        
        ## Hour Angle   
        #HRA = (15.0*(LST-12.0))*np.pi/180.0
        
        ## Declination angle (rad)
        #delta = -23.44*np.cos((360./365.)*(day+10.)*np.pi/180.)*np.pi/180.
        
        ## Zenith angle (rad)
        #psi = np.arccos(np.sin(delta)*np.sin(latitude*np.pi/180.0)+np.cos(delta)*np.cos(latitude*np.pi/180.0)*np.cos(HRA))
        
        ## Solar Azimuth angle, Duffie/Beckman 1.6.6
        #gammas = np.sign(HRA)*np.abs((np.cos(psi)*np.sin(latitude*np.pi/180.)-np.sin(delta))/(np.sin(psi)*np.cos(latitude*np.pi/180)))
        
        ## Slope of the solar panel, Bower AIAA 2011-7072 EQN 15
        #beta = np.reshape(np.arccos(np.cos(thetap)*np.cos(phip)),np.shape(gammas))
        
        ## Angle of incidence, Duffie/Beckman 1.6.3
        #theta = np.arccos(np.cos(psi)*np.cos(beta)+np.sin(psi)*np.sin(beta)*np.cos(gammas-gamma))
        
        #flux = np.zeros_like(psi)
        
        #earth = RCADE.Attributes.Planets.Earth()
        #Re = earth.mean_radius
        
        ## Atmospheric properties
        #Yatm = 9. # The atmospheres thickness in km
        #r    = Re/Yatm
        #c    = altitude/9000.
        
        ## Air mass
        #AM = np.zeros_like(psi)
        #AM[altitude<9000.] = (((r+c[altitude<9000.])*(r+c[altitude<9000.]))*(np.cos(psi[altitude<9000.])*np.cos(psi[altitude<9000.]))+2.*r*(1.-c[altitude<9000.])-c[altitude<9000.]*c[altitude<9000.] +1.)**(0.5)-(r+c[altitude<9000.])*np.cos(psi[altitude<9000.])
        
        ## Direct component 
        #Id = Ind*Io*(0.7**(AM**0.678))
        
        ## Horizontal flux
        #Ih = Id*(np.cos(latitude*np.pi/180.)*np.cos(delta)*np.cos(HRA)+np.sin(latitude*np.pi/180.)*np.sin(delta))              
        
        ## Flux on the inclined panel, if the altitude is less than 9000 meters
        #I = Ih*np.cos(theta)/np.cos(psi)
        
        ## Now update if the plane is outside the majority of the atmosphere, (9km)
        #Id = Ind*Io
        #Ih = Id*(np.cos(latitude*np.pi/180.)*np.cos(delta)*np.cos(HRA)+np.sin(latitude*np.pi/180.)*np.sin(delta))           
        #I[altitude>9000.] = Ih[altitude>9000.]*np.cos(theta[altitude>9000.])/np.cos(psi[altitude>9000.])
        
        ## Now if the sun is on the other side of the Earth...
        #I[((psi<-np.pi/2.)|(psi>96.70995*np.pi/180.))] = 0
        
        #flux = np.maximum(0.0,I)        
        
        ## Store to outputs
        #self.outputs.flux = flux      
        
        ## Return result for fun/convenience
        #return flux
    
    #def power(self):
        #"""This determines the power output of the solar cell.

        #Assumptions:
        #None

        #Source:
        #None

        #Inputs:
        #self.inputs.flux   [W/m^2]

        #Outputs:
        #self.outputs.power [W]
        #power              [W]

        #Properties Used:
        #self.efficiency    [-]
        #self.area          [m^2]
        #"""        
        ## Unpack
        #flux       = self.inputs.flux
        #efficiency = self.efficiency
        #area       = self.area
        
        #p = flux*area*efficiency
        
        ## Store to outputs
        #self.outputs.power = p
    
        #return p
    
    
    
    


# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/write_mass_file.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Library.Methods.Aerodynamics.Athena_Vortex_Lattice.purge_files              import purge_files  

# ----------------------------------------------------------------------------------------------------------------------
#  write_mass_file
# ---------------------------------------------------------------------------------------------------------------------- 
def write_mass_file(avl_object,run_conditions):
    """This function writes the translated aircraft mass into text file read 
    by AVL when it is called
    """    
    
    # unpack inputs
    mass_file       = avl_object.settings.filenames.mass_file 
    aircraft        = avl_object.vehicle
    
    # Open the mass file after purging if it already exists
    purge_files([mass_file]) 
    mass_file_script  = open(mass_file,'w')

    with open(mass_file,'w') as mass_file_script:        
        """This function writes the header using the template required for the AVL executable to read
        """   
        # mass file template
        base_text = \
'''
#-------------------------------------------------
#  {0}
#
#  Dimensional unit and parameter data.
#  Mass & Inertia breakdown.
#-------------------------------------------------

#  Names and scalings for units to be used for trim and eigenmode calculations.
#  The Lunit and Munit values scale the mass, xyz, and inertia table data below.
#  Lunit value will also scale all lengths and areas in the AVL input file.
Lunit = 1.0 m
Munit = 1.0 kg
Tunit = 1.0 s

#------------------------- 
#  Gravity and density to be used as default values in trim setup.
#  Must be in the units given above.
g   = {1}
rho = {2}

#-------------------------
#  Mass & Inertia breakdown.
#  x y z  is location of item's own CG.
#  Ixx... are item's inertias about item's own CG.
#
#  x,y,z system here must be exactly the same one used in the AVL input file
#     (same orientation, same origin location, same length units)
#
#  mass     x     y     z    Ixx    Iyy    Izz   Component Name
#   
    {3}  {4}  {5}  {6}  {7}  {8}  {9} ! {0}
'''

        # Unpack inputs
        name    = avl_object.vehicle.tag
        density = run_conditions.freestream.density[0][0] 
        gravity = run_conditions.freestream.gravity[0][0] 
        
        if aircraft.mass_properties.mass == 0:
            mass = aircraft.mass_properties.max_takeoff
        elif aircraft.mass_properties.max_takeoff == 0:
            mass = aircraft.mass_properties.mass
        else:
            raise AttributeError("Specify Vehicle Mass")
         
        x       = aircraft.mass_properties.center_of_gravity[0][0]
        y       = aircraft.mass_properties.center_of_gravity[0][1]
        z       = aircraft.mass_properties.center_of_gravity[0][2]
        Ixx     = aircraft.mass_properties.moments_of_inertia.tensor[0][0]
        Iyy     = aircraft.mass_properties.moments_of_inertia.tensor[1][1]
        Izz     = aircraft.mass_properties.moments_of_inertia.tensor[2][2]
        
        # Insert inputs into the template
        text = base_text.format(name, gravity , density,mass, x,y,z,Ixx,Iyy,Izz)
        mass_file_script.write(text)  
            
    return


# RCAIDE/Library/Compoments/Wings/Vertical_Tail.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from .Wing import Wing 
from copy import deepcopy
from RCAIDE.Library.Methods.Weights.Moment_of_Inertia.compute_wing_moment_of_inertia import  compute_wing_moment_of_inertia

# ---------------------------------------------------------------------------------------------------------------------- 
#  Vertical_Tail
# ----------------------------------------------------------------------------------------------------------------------  
class Vertical_Tail(Wing):
    """
    A class representing a vertical stabilizer surface for aircraft directional control 
    and stability.

    Attributes
    ----------
    tag : str
        Unique identifier for the vertical tail, defaults to 'vertical_tail'
        
    vertical : bool
        Flag indicating vertical orientation, defaults to True
        
    symmetric : bool
        Flag indicating if tail is symmetric about x-z plane, defaults to False

    Notes
    -----
    The vertical tail provides directional stability and serves as a mounting surface 
    for the rudder. It inherits all geometric and aerodynamic functionality from the 
    Wing class and adds specific attributes for vertical tail operation.

    See Also
    --------
    RCAIDE.Library.Components.Wings.Wing
        Base wing class providing core functionality
    RCAIDE.Library.Components.Wings.Control_Surfaces.Rudder
        Control surface typically mounted on vertical tail
    RCAIDE.Library.Components.Wings.Vertical_Tail_All_Moving
        All-moving vertical tail variant
    """ 

    def __defaults__(self):
        """
        Sets default values for the vertical tail attributes.
        """
        self.tag       = 'vertical_tail'
        self.vertical  = True
        self.symmetric = False 
        
    def make_x_z_reflection(self):
        """
        Creates a reflected copy of the vertical tail over the x-z plane.

        Returns
        -------
        Component
            Reflected vertical tail with appropriate sign conventions

        Notes
        -----
        * Used when vertical tail's symmetric attribute is True
        * Reflects dihedral angles and origin location
        * Control surface deflections are reflected according to sign_duplicate
        * Should be called after setting control surface deflections

        WARNING: this uses deepcopy to achieve its purpose. If this copies too many unwanted 
        attributes, it is recommended that the user should write their own code, taking 
        after the form of this function.
        
        It is also recommended that the user call this function after they set control surface
        or all moving surface deflections. This way the deflection is also properly reflected 
        to the other side
        """    
        wing = deepcopy(self)
        wing.dihedral     *= -1
        wing.origin[0][1] *= -1
        
        for segment in wing.segments:
            segment.dihedral_outboard *= -1
            
        for cs in wing.control_surfaces:
            cs.deflection *= -1*cs.sign_duplicate
                
        return wing
    
    def moment_of_inertia(wing, center_of_gravity):
        """
        Computes the moment of inertia tensor for the vertical tail.

        Parameters
        ----------
        wing : Component
            Wing component data
        center_of_gravity : list
            Reference point coordinates for moment calculation

        Returns
        -------
        ndarray
            3x3 moment of inertia tensor
        """
        I = compute_wing_moment_of_inertia(wing, center_of_gravity) 
        return I      


# RCAIDE/Library/Methods/Propulsors/Converters/Engine/compute_power_from_throttle.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
 # RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import Units

# package imports
import numpy as np 

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_power_from_throttle
# ----------------------------------------------------------------------------------------------------------------------    
def compute_power_from_throttle(engine,engine_conditions,conditions):
    """ The internal combustion engine output power and specific power consumption.
    The following properties are computed:
        engine
          .outputs.power                           (numpy.ndarray): Brake power (or Shaft power)  [W]
          .outputs.power_specific_fuel_consumption (numpy.ndarray): Power (brake) SFC             [lbf/(HP· h) 
          .outputs.fuel_flow_rate                  (numpy.ndarray): Fuel flow                     [kg/s]
          .outputs.torque                          (numpy.ndarray): Torque                        [Q]
    
    Source:
        Gagg and Ferrar model (ref: S. Gudmundsson, 2014 - eq. 7-16)
    
    Assumtions:
        None 
    
    Args:
        engine
          .sea_level_power                        (float): sea-level power      [W]
          .flat_rate_altitude                     (float): flat rate altitude   [m]
          .power_specific_fuel_consumption        (float): PSFC                 [RPM]
          .throttle setting               (numpy.ndarray): throttle             [unitless]
          .inputs.speed                   (numpy.ndarray): angular velocity     [rad/s]
        freestream conditions 
          .altitude                               (float): altitde              [m]
          .delta_isa                              (float): altitude offset      [m]
            
    Returns:
        None 
    """

    # Unpack
    altitude    = conditions.freestream.altitude
    delta_isa   = conditions.freestream.delta_ISA 
    PSLS        = engine.sea_level_power  # SI Units 
    h_flat      = engine.flat_rate_altitude
    omega       = engine_conditions.speed
    PSFC        = engine.power_specific_fuel_consumption  

    # shift in power lapse due to flat rate
    altitude_virtual = altitude - h_flat       
    altitude_virtual[altitude_virtual<0.] = 0. 

    # Compute the sea-level ISA atmosphere conditions
    atmo             = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_values_0    = atmo.compute_values(0,0) 
    rho0             = atmo_values_0.density[0,0]     
    atmo_values      = atmo.compute_values(altitude_virtual,delta_isa) 
    rho              = atmo_values.density 
    
    # Compute the density ratio:
    sigma = rho / rho0
    
    # Compute available power 
    Pavailable                    = PSLS * (sigma - 0.117) / 0.883        
    Pavailable[h_flat > altitude] = PSLS

    # Regulate using throttle 
    P       = Pavailable * engine_conditions.throttle 
    P[P<0.] = 0.
    
    m_dot  =  PSFC * P 

    # Compute engine torque
    torque = P/omega
    
    # Determine fuel flow rate and cap at 0
    fuel_flow_rate  = np.fmax(m_dot,np.zeros_like(altitude)) 
    
    # Store results 
    engine_conditions.power                           = P
    engine_conditions.power_specific_fuel_consumption = PSFC
    engine_conditions.fuel_flow_rate                  = fuel_flow_rate
    engine_conditions.torque                          = torque

    return



# RCAIDE/Library/Plots/Emissions/plot_CO2e_emissions
# 
# 
# Created:  Jul 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------    
def plot_CO2e_emissions(results,
                             save_figure = False,
                             show_legend = True,
                             save_filename = "CO2e_Emissions" ,
                             file_type = ".png",
                             width = 11, height = 7):
    """  
 
    """
 
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
      
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height) 
    axis_1 = plt.subplot(1,1,1)

    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))
    
    cum_y0  = 0
    cum_y1  = 0 
    cum_y1_0 = 0  
    
    for i in range(len(results.segments)): 
        time                = results.segments[i].conditions.frames.inertial.time[:, 0] / Units.min 
        emissions_CO2       = results.segments[i].conditions.emissions.total.CO2[:, 0]  
        emissions_NOx       = results.segments[i].conditions.emissions.total.NOx[:, 0] 
        emissions_H2O       = results.segments[i].conditions.emissions.total.H2O[:, 0] 
        emissions_Contrails = results.segments[i].conditions.emissions.total.Contrails[:, 0]
        emissions_Soot      = results.segments[i].conditions.emissions.total.Soot[:, 0]
        emissions_SO2       = results.segments[i].conditions.emissions.total.SO2[:, 0]  

        cum_y0 = np.zeros_like(emissions_CO2)  
        cum_y1 = cum_y1_0 + emissions_CO2 +  emissions_NOx  + emissions_H2O  + emissions_Contrails +  emissions_Soot +  emissions_SO2    

        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ')    
        axis_1.fill_between(time, cum_y0, cum_y1, where=(cum_y0 < cum_y1), color= line_colors[i],  interpolate=True, label = segment_name)   
        cum_y1_0 = cum_y1[-1] 
                
        axis_1.set_ylabel(r'CO2e Emissions (kg)') 
        set_axes(axis_1)
        
    if show_legend:
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})    
    
    # Adjusting the sub-plots for legend 
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'CO2e Emissions'      
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return fig 


# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/write_avl_airfoil_file.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Library.Methods.Aerodynamics.Athena_Vortex_Lattice.purge_files  import purge_files
from RCAIDE.Library.Methods.Geometry.Airfoil.import_airfoil_geometry        import import_airfoil_geometry 

# python imports
import os

# ----------------------------------------------------------------------------------------------------------------------
#  write_avl_airfoil_file
# ---------------------------------------------------------------------------------------------------------------------- 
def write_avl_airfoil_file(rcaide_airfoil_filename):
    """ This function writes the standard airfoil file format from Airfoil tools 
    to avl file format

    Assumptions:
        None
        
    Source: 
   
    Inputs:
        filename
   
    Outputs:
        airfoil.dat
   
    Properties Used:
        N/A
    """       
    

    # unpack avl_inputs
    avl_airfoil_filename =  rcaide_airfoil_filename.split(".")[-2].split("/")[-1] + '.dat'
    
    # purge file 
    purge_files([avl_airfoil_filename]) 
    
    # read airfoil file header 
    origin  = os.getcwd()
    os_path = os.path.split(origin)[0]
    f_path  = rcaide_airfoil_filename
    f = open(f_path)  
    data_block = f.readlines() 
    f.close()   
    airfoil_name = data_block[0].strip()
    
    # import airfoil coordinates 
    airfoil_geometry_data = import_airfoil_geometry(f_path)
    dim = len(airfoil_geometry_data.x_coordinates)
              
    # write file  
    with open(avl_airfoil_filename,'w') as afile:
            afile.write(airfoil_name + "\n")  
            for i in range(dim - 1):
                if i == int(dim/2):
                    pass  
                elif airfoil_geometry_data.y_coordinates[i] < 0.0:
                    case_text = '\t' + format(airfoil_geometry_data.x_coordinates[i], '.7f')+ "   " + format(airfoil_geometry_data.y_coordinates[i], '.7f') + "\n" 
                    afile.write(case_text)
                else:   
                    case_text = '\t' + format(airfoil_geometry_data.x_coordinates[i], '.7f')+ "    " + format(airfoil_geometry_data.y_coordinates[i], '.7f') + "\n" 
                    afile.write(case_text)
    afile.close()
    return avl_airfoil_filename 
 



## @ingroup Core
# DiffedData.py
#
# Created:  Feb 2015, T. Lukacyzk
# Modified: Feb 2016, T. MacDonald
#           Jun 2016, E. Botero

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

from copy import deepcopy
from .Container import Container as ContainerBase
from .Data import Data
from .DataOrdered import DataOrdered
import numpy as np

# ----------------------------------------------------------------------
#  Config
# ----------------------------------------------------------------------

## @ingroup Core
class Diffed_Data(Data):
    """ This is for creating a data new class where a different copy is saved.
        This is useful for creating a new configuration of a vehicle.

        Assumptions:
        N/A

        Source:
        N/A
    """    

    
    def __defaults__(self):
        """ A stub for all classes that come later
            
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """  
        self.tag    = 'config'
        self._base  = Data()
        self._diff  = Data()
        
    def __init__(self,base=None):
        """ Initializes the new Diffed_Data() class through a deepcopy
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """  
        if base is None: base = Data()
        self._base = base
        this = deepcopy(base) # deepcopy is needed here to build configs - Feb 2016, T. MacDonald
        Data.__init__(self,this)
        
    def store_diff(self):
        """ Finds the differences and saves them
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        delta = diff(self,self._base)
        self._diff = delta
         
# ----------------------------------------------------------------------
#  Config Container
# ----------------------------------------------------------------------

class Container(ContainerBase):
    """ A dict-type container with attribute, item and index style access
        intended to hold a attribute-accessible list of Data(). This is unordered.
        
        Assumptions:
        N/A
        
        Source:
        N/A
        
    """
    def append(self,value):
        """ Appends the value to the containers
        
            Assumptions:
            None
        
            Source:
            N/A
        
            Inputs:
            self
        
            Outputs:
            N/A
            
            Properties Used:
            N/A
        """         
        try: value.store_diff()
        except AttributeError: pass
        ContainerBase.append(self,value)
        
    def pull_base(self):
        """ Updates the differences
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        for config in self:
            try: config.pull_base()
            except AttributeError: pass

    def store_diff(self):
        """ Finds the differences and saves them
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """          
        for config in self:
            try: config.store_diff()
            except AttributeError: pass
    
    def finalize(self):
        """ This just does a pull_base()
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            N/A
    
            Outputs:
            N/A
    
            Properties Used:
            N/A    
        """        
        for config in self:
            try: config.finalize()
            except AttributeError: pass


# ------------------------------------------------------------
#  Handle Linking
# ------------------------------------------------------------

Diffed_Data.Container = Container

# ------------------------------------------------------------
#  Diffing Function
# ------------------------------------------------------------

def diff(A,B):
    """ The magic diff function that makes Diffed_Data() work

        Assumptions:
        N/A

        Source:
        N/A

        Inputs:
        A
        B

        Outputs:
        Result

        Properties Used:
        N/A    
    """      

    keys = set([])
    keys.update( A.keys() )
    keys.update( B.keys() )

    if isinstance(A,Diffed_Data):
        keys.remove('_base')
        keys.remove('_diff')

    result = type(A)()
    result.clear()

    for key in keys:
        va = A.get(key,None)
        vb = B.get(key,None)
        if isinstance(va,Data) and isinstance(vb,Data):
            sub_diff = diff(va,vb)
            if sub_diff:
                result[key] = sub_diff

        elif isinstance(va,Data) or isinstance(vb,Data):
            result[key] = va
            
        elif isinstance(va,DataOrdered) and isinstance(vb,DataOrdered):
            sub_diff = diff(va,vb)
            if sub_diff:
                result[key] = sub_diff

        elif isinstance(va,DataOrdered) or isinstance(vb,DataOrdered):
            result[key] = va        

        elif not np.all(va == vb):
            result[key] = va

    return result    


# RCAIDE/Methods/Noise/Common/generate_microphone_locations.py
# 
# 
# Created:  Oct 2023, A. Molloy  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# package imports 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  generate_hemisphere_microphone_locations
# ---------------------------------------------------------------------------------------------------------------------- 
def generate_hemisphere_microphone_locations(settings): 
    """This computes the microphones locations in a noise hemisphere
            
    Assumptions:
        None

    Source:
        N/A  

    Inputs:   
    settings.
    
        r     - noise hemisphere radius                   [meters]
        n     - noise hemisphere microphone resolution    [unitless]
        phi   - noise hemisphere phi angle bounds         [radians]
        theta - noise hemisphere theta angle bounds       [radians] 
    
    Outputs: 
        gm_mic_locations   - cartesian coordiates of all microphones defined  [meters] 
    
    Properties Used:
        N/A       
    """     
    r     = settings.noise_hemisphere_radius                  
    phi   = settings.noise_hemisphere_phi_angles   
    theta = settings.noise_hemisphere_theta_angles   
 
    x     = r * np.outer(np.sin(phi), np.cos(theta))
    y     = r * np.outer(np.sin(phi), np.sin(theta))
    z     = r * np.outer(np.cos(phi), np.ones(np.size(theta))) 
 
    num_gm                = len(z.flatten())
    gm_mic_locations      = np.zeros((num_gm,3))  
    gm_mic_locations[:,0] = x.flatten() 
    gm_mic_locations[:,1] = y.flatten() 
    gm_mic_locations[:,2] = z.flatten()
    
    return gm_mic_locations   


# RCAIDE/Library/Components/Energy/Sources/Solar_Panels/__init__.py
# 

"""
Solar panel components for aircraft energy generation systems

This module provides implementations for solar panel systems used in aircraft
power generation. It includes models for photovoltaic cells, maximum power point
tracking, and integration with aircraft electrical systems.

See Also
--------
RCAIDE.Library.Components.Energy.Modulators.Solar_Logic
    Solar power management and control components
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Solar_Panel        import Solar_Panel


## @defgroup Methods-Propulsion-Rotor_Wake-Fidelity_Zero
# compute_fidelity_zero_induced_velocity.py
# 
# Created:  Jun 2021, R. Erhard 

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# package imports
import numpy as np 
from scipy.interpolate import interp1d

## @defgroup Methods-Propulsion-Rotor_Wake-Fidelity_Zero
def compute_fidelity_zero_induced_velocity(rotor,rotor_conditions,evaluation_points,ctrl_pts, identical_flag=False):  
    """ This computes the velocity induced by the fidelity zero wake
    on specified evaluation points.

    Assumptions:  
       The wake contracts following formulation by McCormick.
    
    Source:   
       Contraction factor: McCormick 1969, Aerodynamics of V/STOL Flight
       
    Inputs: 
       evaluation_points.
          XC              - X-location of evaluation points (vehicle frame)             [m] 
          YC              - Y-location of evaluation points (vehicle frame)             [m] 
          ZC              - Z-location of evaluation points (vehicle frame)             [m] 
          
       geometry    - SUAVE vehicle                                 [Unitless] 
       cpts        - control points in segment                     [Unitless]
    Properties Used:
       N/A
    """

    # extract vortex distribution
    n_cp = len(evaluation_points.XC)
    
    # initialize rotor wake induced velocities
    rotor_V_wake_ind = np.zeros((ctrl_pts,n_cp,3))
       
    R            = rotor.tip_radius
    r            = rotor_conditions.disc_radial_distribution[0,:,0]
    
    # Ignore points within hub or outside tip radius
    hub_y_center = rotor.origin[0][1]
    inboard_r    = np.flip(hub_y_center - r) 
    outboard_r   = hub_y_center + r 
    rotor_y_range = np.append(inboard_r, outboard_r)

    # within this range, add an induced x- and z- velocity from rotor wake
    bool_inboard  = ( evaluation_points.YC > inboard_r[0] )  * ( evaluation_points.YC < inboard_r[-1] )
    bool_outboard = ( evaluation_points.YC > outboard_r[0] ) * ( evaluation_points.YC < outboard_r[-1] )
    bool_in_range = bool_inboard + bool_outboard
    YC_in_range   = evaluation_points.YC[bool_in_range]

    y_vals  = YC_in_range
    val_ids = np.where(bool_in_range==True)
    
    s  = evaluation_points.XC[val_ids] - rotor.origin[0][0]
    kd = 1 + s/(np.sqrt(s**2 + R**2))    

    # extract radial and azimuthal velocities at blade
    va = rotor_conditions.blade_axial_induced_velocity[0]
    vt = rotor_conditions.blade_tangential_induced_velocity[0]
    
    
    va_y_range  = np.append(np.flipud(va), va)
    vt_y_range  = np.append(np.flipud(vt), vt)*rotor.rotation
    va_interp   = interp1d(rotor_y_range, va_y_range)
    vt_interp   = interp1d(rotor_y_range, vt_y_range)
    
    
    # preallocate va_new and vt_new
    va_new = kd*va_interp((y_vals))
    vt_new = np.zeros(np.size(val_ids))
    
    # invert inboard vt values
    inboard_bools                = (y_vals < hub_y_center)
    vt_new[inboard_bools]        = -kd[inboard_bools]*vt_interp((y_vals[inboard_bools]))
    vt_new[inboard_bools==False] = kd[inboard_bools==False]*vt_interp((y_vals[inboard_bools==False]))
    
    rotor_V_wake_ind[0,val_ids,0] = va_new  # axial induced velocity
    rotor_V_wake_ind[0,val_ids,1] = 0       # spanwise induced velocity; in line with rotor, so 0
    rotor_V_wake_ind[0,val_ids,2] = vt_new  # vertical induced velocity     
    
    return rotor_V_wake_ind
  
  


# RCAIDE/Methods/Energy/Propulsors/Rotor_Design/optimization_setup.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
 
# RCAIDE Imports  
import RCAIDE 
from RCAIDE.Framework.Core                                                                    import Units, Data   
from RCAIDE.Framework.Optimization.Common                                                     import Nexus       
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Design.blade_geometry_setup    import blade_geometry_setup
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Design.procedure_setup         import procedure_setup

# Python package imports   
import numpy as np  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Optimization Setuo 
# ----------------------------------------------------------------------------------------------------------------------    
def optimization_setup(rotor,number_of_stations,print_iterations):
    """ Sets up rotor optimization problem including design variables, constraints and objective function
        using RCAIDE's Nexus optimization framework. Appends methodolody of planform modification to Nexus.
          Inputs: 
             rotor     - rotor data structure           [None]
             
          Outputs: 
              nexus    - RCAIDE's optimization framework [None]
              
          Assumptions: 
            1) minimum allowable blade taper : 0.2  
            1) maximum allowable blade taper : 0.7     
        
          Source:
             None
    """    
    nexus                        = Nexus()
    problem                      = Data()
    nexus.optimization_problem   = problem
   
    if type(rotor) != RCAIDE.Library.Components.Propulsors.Converters.Prop_Rotor or  type(rotor) != RCAIDE.Library.Components.Propulsors.Converters.Lift_Rotor:
        assert('rotor must be of Lift-Rotor or Prop-Rotor class') 
        
    if type(rotor) == RCAIDE.Library.Components.Propulsors.Converters.Prop_Rotor:
        nexus.prop_rotor_flag = True 
    else:
        nexus.prop_rotor_flag = False 
    # -------------------------------------------------------------------
    # Inputs
    # -------------------------------------------------------------------  
    R         = rotor.tip_radius  
    tm_ll_h   = rotor.optimization_parameters.tip_mach_range[0]
    tm_ul_h   = rotor.optimization_parameters.tip_mach_range[1] 
    tm_0_h    = (tm_ul_h + tm_ll_h)/2 
    
    if nexus.prop_rotor_flag:    
        tm_ll_c          = rotor.optimization_parameters.tip_mach_range[0]
        tm_ul_c          = rotor.optimization_parameters.tip_mach_range[1]    
        
    inputs = []
    inputs.append([ 'chord_r'               ,  0.1*R    , 0.05*R     , 0.2*R     , 1.0     ,  1*Units.less])
    inputs.append([ 'chord_p'               ,  2        , 0.25       , 2.0       , 1.0     ,  1*Units.less])
    inputs.append([ 'chord_q'               ,  1        , 0.25       , 1.5       , 1.0     ,  1*Units.less])
    inputs.append([ 'chord_t'               ,  0.05*R   , 0.02*R     , 0.1*R     , 1.0     ,  1*Units.less])  
    inputs.append([ 'twist_r'               ,  np.pi/6  ,  0         , np.pi/4   , 1.0     ,  1*Units.less])
    inputs.append([ 'twist_p'               ,  1        , 0.25       , 2.0       , 1.0     ,  1*Units.less])
    inputs.append([ 'twist_q'               ,  0.5      , 0.25       , 1.5       , 1.0     ,  1*Units.less])
    inputs.append([ 'twist_t'               ,  np.pi/6  , 0          , np.pi/4   , 1.0     ,  1*Units.less])  
    inputs.append([ 'hover_tip_mach'        , tm_0_h    , tm_ll_h    , tm_ul_h   , 1.0     ,  1*Units.less])
    inputs.append([ 'OEI_tip_mach'          , tm_0_h    , tm_ll_h    , 0.85      , 1.0     ,  1*Units.less])  
    inputs.append([ 'OEI_collective_pitch'  , np.pi/6   , -np.pi/6   , np.pi/6   , 1.0      ,  1*Units.less]) 
    if nexus.prop_rotor_flag: 
        inputs.append([ 'cruise_tip_mach'         , tm_ll_c , tm_ll_c    , tm_ul_c   , 1.0     ,  1*Units.less]) 
        inputs.append([ 'cuise_collective_pitch'  ,-np.pi/6  , -np.pi/6   , np.pi/6  , 1.0     ,  1*Units.less]) 
    problem.inputs = np.array(inputs,dtype=object)   

    # -------------------------------------------------------------------
    # Objective
    # ------------------------------------------------------------------- 
    problem.objective = np.array([  
                                 [  'objective'  ,  1.0   ,    1*Units.less] 
    ],dtype=object)
    
    # -------------------------------------------------------------------
    # Constraints
    # -------------------------------------------------------------------  
    constraints = [] 
    constraints.append([ 'hover_thrust_pow_res'      ,  '<'  ,  1E-3 ,   1.0   , 1*Units.less])  
    constraints.append([ 'blade_taper_constraint_1'  ,  '>'  ,  0.3 ,   1.0   , 1*Units.less])  
    constraints.append([ 'blade_taper_constraint_2'  ,  '<'  ,  0.9 ,   1.0   , 1*Units.less])  
    constraints.append([ 'blade_twist_constraint'    ,  '>'  ,  0.0 ,   1.0   , 1*Units.less])
    constraints.append([ 'max_sectional_cl_hov'      ,  '<'  ,  0.8 ,   1.0   , 1*Units.less])
    constraints.append([ 'chord_p_to_q_ratio'        ,  '>'  ,  0.5 ,   1.0   , 1*Units.less])    
    constraints.append([ 'twist_p_to_q_ratio'        ,  '>'  ,  0.5 ,   1.0   , 1*Units.less])  
    constraints.append([ 'OEI_hov_thrust_pow_res'    ,  '<'  ,  1E-3 ,   1.0   , 1*Units.less]) 
    if nexus.prop_rotor_flag:
        constraints.append([ 'cruise_thrust_pow_res'     ,  '<'  ,  1E-3 ,   1.0   , 1*Units.less]) 
        constraints.append([ 'max_sectional_cl_cruise'   ,  '<'  ,  0.8 ,   1.0   , 1*Units.less])   
    problem.constraints =  np.array(constraints,dtype=object)                
    
    # -------------------------------------------------------------------
    #  Aliases
    # ------------------------------------------------------------------- 
    aliases = [] 
    aliases.append([ 'chord_r'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.chord_r' ])
    aliases.append([ 'chord_p'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.chord_p' ])
    aliases.append([ 'chord_q'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.chord_q' ])
    aliases.append([ 'chord_t'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.chord_t' ]) 
    aliases.append([ 'twist_r'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.twist_r' ])
    aliases.append([ 'twist_p'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.twist_p' ])
    aliases.append([ 'twist_q'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.twist_q' ])
    aliases.append([ 'twist_t'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.twist_t' ])     
    aliases.append([ 'hover_tip_mach'             , 'vehicle_configurations.hover.networks.electric.propulsors.electric_rotor.rotor.hover.design_tip_mach' ]) 
    aliases.append([ 'objective'                  , 'summary.objective'       ])  
    aliases.append([ 'hover_thrust_pow_res'       , 'summary.hover_thrust_power_residual'   ]) 
    aliases.append([ 'blade_taper_constraint_1'   , 'summary.blade_taper_constraint_1'])   
    aliases.append([ 'blade_taper_constraint_2'   , 'summary.blade_taper_constraint_2'])   
    aliases.append([ 'blade_twist_constraint'     , 'summary.blade_twist_constraint'])    
    aliases.append([ 'max_sectional_cl_hov'       , 'summary.max_sectional_cl_hover'])   
    aliases.append([ 'chord_p_to_q_ratio'         , 'summary.chord_p_to_q_ratio'    ])  
    aliases.append([ 'twist_p_to_q_ratio'         , 'summary.twist_p_to_q_ratio'    ])   
    aliases.append([ 'OEI_hov_thrust_pow_res'     , 'summary.oei_thrust_power_residual'   ]) 
    aliases.append([ 'OEI_collective_pitch'       , 'vehicle_configurations.oei.networks.electric.propulsors.electric_rotor.rotor.oei.design_pitch_command' ]) 
    aliases.append([ 'OEI_tip_mach'               , 'vehicle_configurations.oei.networks.electric.propulsors.electric_rotor.rotor.oei.design_tip_mach' ]) 
    if nexus.prop_rotor_flag: 
        aliases.append([ 'cruise_tip_mach'        , 'vehicle_configurations.cruise.networks.electric.propulsors.electric_rotor.rotor.cruise.design_tip_mach' ])  
        aliases.append([ 'cuise_collective_pitch' , 'vehicle_configurations.cruise.networks.electric.propulsors.electric_rotor.rotor.cruise.design_pitch_command' ])  
        aliases.append([ 'cruise_thrust_pow_res'  , 'summary.cruise_thrust_power_residual'   ]) 
        aliases.append([ 'max_sectional_cl_cruise', 'summary.max_sectional_cl_cruise'])  
         
    problem.aliases = aliases
    
    # -------------------------------------------------------------------
    #  Vehicles
    # -------------------------------------------------------------------
    nexus.vehicle_configurations = blade_geometry_setup(rotor,number_of_stations)
    
    # -------------------------------------------------------------------
    #  Analyses
    # -------------------------------------------------------------------
    nexus.analyses = None 
    
    # -------------------------------------------------------------------
    #  Missions
    # -------------------------------------------------------------------
    nexus.missions = None
    
    # -------------------------------------------------------------------
    #  Procedure
    # -------------------------------------------------------------------    
    nexus.print_iterations  = print_iterations 
    nexus.procedure         = procedure_setup()
    
    # -------------------------------------------------------------------
    #  Summary
    # -------------------------------------------------------------------    
    nexus.summary        = Data()     
    nexus.results.hover  = Data() 
    nexus.results.cruise = Data()
    nexus.results.oei    = Data()
    
    return nexus   


# RCAIDE/Compoments/Wings/Control_Surfaces/Control_Surface.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports    
from RCAIDE.Library.Components import Component 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Control Surface
# ---------------------------------------------------------------------------------------------------------------------- 
class Control_Surface(Component):
    """
    Base class for aircraft control surfaces providing core functionality for geometric 
    definition and operation.

    Attributes
    ----------
    tag : str
        Unique identifier for the control surface, defaults to 'control_surface'
        
    span : float
        Spanwise length of the control surface in meters, defaults to 0.0
        
    span_fraction_start : float
        Spanwise location where control surface begins as fraction of wing span, 
        defaults to 0.0
        
    span_fraction_end : float
        Spanwise location where control surface ends as fraction of wing span, 
        defaults to 0.0
        
    hinge_fraction : float
        Chordwise location of hinge line as fraction of control surface chord, 
        defaults to 0.0
        
    chord_fraction : float
        Fraction of wing chord occupied by control surface, defaults to 0.0
        
    sign_duplicate : float
        Sign convention for duplicate control surface deflection (1.0 or -1.0), 
        defaults to 1.0
        
    deflection : float
        Control surface deflection angle, defaults to 0.0
        
    configuration_type : str
        Type of control surface construction (e.g., 'single_slotted'), 
        defaults to 'single_slotted'
        
    gain : float
        Deflection multiplier used for AVL analysis, defaults to 1.0

    Notes
    -----
    The control surface class provides the foundation for all movable aerodynamic 
    surfaces. Key features include:
    
    * Geometric definition relative to parent wing
    * Deflection and hinge line specification
    * Configuration type specification
    * Duplicate surface handling

    **Definitions**

    'Span Fraction'
        Location along wing span measured as fraction from root (0.0) to tip (1.0)
        
    'Chord Fraction'
        Portion of local wing chord occupied by control surface
        
    'Hinge Fraction'
        Location of hinge line measured as fraction of control surface chord
        
    'Sign Duplicate'
        Determines deflection relationship between paired surfaces:
        * 1.0 for synchronized deflection (e.g., flaps)
        * -1.0 for opposite deflection (e.g., ailerons)

    See Also
    --------
    RCAIDE.Library.Components.Wings.Control_Surfaces.Flap
        High-lift device implementation
    RCAIDE.Library.Components.Wings.Control_Surfaces.Aileron
        Roll control implementation
    RCAIDE.Library.Components.Wings.Control_Surfaces.Elevator
        Pitch control implementation
    RCAIDE.Library.Components.Wings.Control_Surfaces.Rudder
        Yaw control implementation
    """

    def __defaults__(self):
        """
        Sets default values for the control surface attributes.
        """
        self.tag                   = 'control_surface' 
        self.span                  = 0.0
        self.span_fraction_start   = 0.0
        self.span_fraction_end     = 0.0
        
        self.hinge_fraction        = 0.0
        self.chord_fraction        = 0.0
        
        self.sign_duplicate        = 1.0
        self.deflection            = 0.0  
        self.configuration_type    = 'single_slotted'
        
        self.gain                  = 1.0 #deflection multiplier used only for AVL




# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
import numpy as np
from RCAIDE.Library.Methods.Geometry.Planform.convert_sweep import convert_sweep

# ----------------------------------------------------------------------
#  Method
# ---------------------------------------------------------------------- 
def estimate_wing_CL_alpha(wing,mach):
    """ This method uses the DATCOM formula to compute dCL/dalpha without 
    correlations for downwash of lifting surfaces further ahead on the 
    aircraft or upwash resulting from the position of the wing on the body.

    CAUTION: The method presented here is applicable for subsonic speeds.
    May be inaccurate for transonic or supersonic flight. A correction factor
    for supersonic flight is included, but may not be completely accurate.

    Assumptions:
    Mach number should not be transonic
    
    Source:
        None
         
    Inputs:
        wing - a data dictionary with the fields:
            effective_apsect_ratio - wing aspect ratio [dimensionless]. If 
            this variable is not inlcuded in the input, the method will look
            for a variable named 'aspect_ratio'.
            sweep_le - wing leading-edge sweep angle [radians]
            taper - wing taper ratio [dimensionless]
        mach - flight Mach number [dimensionless]. Should be a numpy array
            with one or more elements.

    Outputs:
        cL_alpha - The derivative of 3D lift coefficient with respect to AoA

    Properties Used:
    N/A
    """         
    
    #Unpack inputs
    if 'effective_aspect_ratio' in wing:
        ar = wing.effective_aspect_ratio
    elif 'extended' in wing:
        if 'aspect_ratio' in wing.extended:
            ar = wing.extended.aspect_ratio
        else:
            ar = wing.aspect_ratio
    else:
        ar = wing.aspect_ratio    
        
    #Compute relevent parameters
    cL_alpha = []
    half_chord_sweep = convert_sweep(wing,0.25,0.5)  #Assumes original sweep is that of LE
    
    #Compute k correction factor for Mach number    
    #First, compute corrected 2D section lift curve slope (C_la) for the given Mach number
    cla = 6.13          #Section C_la at M = 0; Roskam Airplane Design Part VI, Table 8.1  
    
    cL_alpha = np.ones_like(mach)
    Beta     = np.ones_like(mach)
    k        = np.ones_like(mach)
    cla_M    = np.ones_like(mach)
    
    Beta[mach<1.]  = (1.0-mach[mach<1.]**2.0)**0.5
    Beta[mach>1.]  = (mach[mach>1.]**2.0-1.0)**0.5
    cla_M[mach<1.] = cla/Beta[mach<1.]
    cla_M[mach>1.] = 4.0/Beta[mach>1.]
    k              = cla_M/(2.0*np.pi/Beta)
    
    #Compute aerodynamic surface 3D lift curve slope using the DATCOM formula
    cL_alpha =(2.0*np.pi*ar/(2.0+((ar**2.0*(Beta*Beta)/(k*k))*(1.0+(np.tan(half_chord_sweep))**2.0/(Beta*Beta))+4.0)**0.5))
    
    return cL_alpha



# RCAIDE/Library/Components/Propulsors/Turboprop.py 
#
#
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports   
from .                          import Propulsor
from RCAIDE.Library.Methods.Propulsors.Turboprop_Propulsor.append_turboprop_conditions     import append_turboprop_conditions 
from RCAIDE.Library.Methods.Propulsors.Turboprop_Propulsor.compute_turboprop_performance   import compute_turboprop_performance, reuse_stored_turboprop_data
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Fan Component
# ---------------------------------------------------------------------------------------------------------------------- 
class Turboprop(Propulsor):
    """
    A turboprop propulsion system model that simulates the performance of a turboprop engine.

    Attributes
    ----------
    tag : str
        Identifier for the turboprop engine. Default is 'turboprop'.
    
    nacelle : Component
        Nacelle component of the engine. Default is None.
        
    compressor : Component
        Compressor component of the engine. Default is None.
        
    turbine : Component
        Turbine component of the engine. Default is None.
        
    combustor : Component
        Combustor component of the engine. Default is None.
        
    active_fuel_tanks : list
        List with names of active fuel tanks. Default is None.
        
    engine_diameter : float
        Diameter of the engine [m]. Default is 0.0.
        
    engine_length : float
        Length of the engine [m]. Default is 0.0.
        
    engine_height : float
        Engine centerline height above the ground plane [m]. Default is 0.5.
        
    design_isa_deviation : float
        ISA temperature deviation at design point [K]. Default is 0.0.
        
    design_altitude : float
        Design altitude of the engine [m]. Default is 0.0.
        
    design_propeller_efficiency : float
        Design point propeller efficiency. Default is 0.0.
        
    design_gearbox_efficiency : float
        Design point gearbox efficiency. Default is 0.0.
        
    design_mach_number : float
        Design Mach number. Default is 0.0.
        
    compressor_nondimensional_massflow : float
        Non-dimensional mass flow through the compressor. Default is 0.0.
        
    reference_temperature : float
        Reference temperature for calculations [K]. Default is 288.15.
        
    reference_pressure : float
        Reference pressure for calculations [Pa]. Default is 101325.0.

    Notes
    -----
    The Turboprop class inherits from the Propulsor base class and implements
    methods for computing turboprop engine performance. A turboprop engine uses
    a gas turbine core to drive a propeller through a reduction gearbox, combining
    the efficiency of a propeller at low speeds with the power of a turbine engine.

    **Definitions**

    'ISA'
        International Standard Atmosphere - standard atmospheric model

    'Mach number'
        Ratio of flow velocity to the local speed of sound

    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Propulsor
    RCAIDE.Library.Components.Propulsors.Turbofan
    RCAIDE.Library.Components.Propulsors.Turbojet
    """ 
    def __defaults__(self):    
        # setting the default values
        self.tag                                      = 'turboprop'   
        self.nacelle                                  = None 
        self.compressor                               = None  
        self.turbine                                  = None  
        self.combustor                                = None  
        self.active_fuel_tanks                        = None         
        self.engine_diameter                          = 0.0      
        self.engine_length                            = 0.0
        self.engine_height                            = 0.5      
        self.design_isa_deviation                     = 0.0
        self.design_altitude                          = 0.0
        self.design_propeller_efficiency              = 0.0
        self.design_gearbox_efficiency                = 0.0 
        self.design_mach_number                       = 0.0
        self.compressor_nondimensional_massflow       = 0.0
        self.reference_temperature                    = 288.15
        self.reference_pressure                       = 1.01325*10**5  
    
    def append_operating_conditions(self,segment):
        """
        Appends operating conditions to the segment.
        """
        append_turboprop_conditions(self,segment)
        return

    def unpack_propulsor_unknowns(self,segment):   
        return 

    def pack_propulsor_residuals(self,segment): 
        return

    def append_propulsor_unknowns_and_residuals(self,segment): 
        return    
    
    def compute_performance(self,state,center_of_gravity = [[0, 0, 0]]):
        """
        Computes turboprop performance including thrust, moment, and power.
        """
        thrust,moment,power,stored_results_flag,stored_propulsor_tag =  compute_turboprop_performance(self,state,center_of_gravity)
        return thrust,moment,power,stored_results_flag,stored_propulsor_tag
    
    def reuse_stored_data(turboprop,state,network,stored_propulsor_tag,center_of_gravity = [[0, 0, 0]]):
        """
        Reuses stored turboprop data for performance calculations.
        """
        thrust,moment,power  = reuse_stored_turboprop_data(turboprop,state,network,stored_propulsor_tag,center_of_gravity)
        return thrust,moment,power 


# RCAIDE/Framework/Analyses/Mission/Segments/Conditions/Results.py
# 
# 
# Created:  Jul 2023, M. Clarke
# Updated:  May 2024, M. Guidotti

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Framework.Mission.Common import Conditions

# python imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# Results
# ---------------------------------------------------------------------------------------------------------------------- 

class Results(Conditions):
    """ This builds upon Basic, which itself builds on conditions, to add the data structure for aerodynamic mission analyses.
    
        Assumptions:
        None
        
        Source:ty.static.coeffici
        None   ty.static.coeffici
    """
    
    
    def __defaults__(self):
        """This sets the default values.
    
            Assumptions:
            Coefficient subscritps:
                X      - force in X direction
                Y      - force in Y direction 
                Z      - force in Z direction
                
                lift  - force 
                drag  - force 
    
                L     - moment about X axis 
                M     - moment about Y axis 
                N     - moment about Z axis
                
                u     - velocity in X drection
                v     - velocity in Y drection
                w     - velocity in Z drection
                   
                p     - angular rate about X axis
                q     - angular rate about Y axis
                r     - angular rate about Z axis
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """ 
        
        self.tag                                              = 'results' 
 
        # start default row vectors
        ones_1col                                             = self.ones_row(1)  
        ones_3col                                             = self.ones_row(3)
        
        # ----------------------------------------------------------------------------------------------------------------------         
        # Reference Values 
        # ---------------------------------------------------------------------------------------------------------------------- 
    
        self.S_ref                                            = ones_1col * 0
        self.c_ref                                            = ones_1col * 0
        self.b_ref                                            = ones_1col * 0
        self.X_ref                                            = ones_1col * 0
        self.Y_ref                                            = ones_1col * 0
        self.Z_ref                                            = ones_1col * 0     
        
        # ----------------------------------------------------------------------------------------------------------------------         
        # Frames 
        # ---------------------------------------------------------------------------------------------------------------------- 
        self.frames                                           = Conditions()
        
        # inertial conditions
        self.frames.inertial                                                   = Conditions()        
        self.frames.inertial.position_vector                                   = ones_3col * 0
        self.frames.inertial.velocity_vector                                   = ones_3col * 0
        self.frames.inertial.acceleration_vector                               = ones_3col * 0
        self.frames.inertial.angular_velocity_vector                           = ones_3col * 0
        self.frames.inertial.angular_acceleration_vector                       = ones_3col * 0
        self.frames.inertial.gravity_force_vector                              = ones_3col * 0
        self.frames.inertial.total_force_vector                                = ones_3col * 0
        self.frames.inertial.total_moment_vector                               = ones_3col * 0
        self.frames.inertial.time                                              = ones_1col * 0
        self.frames.inertial.aircraft_range                                    = ones_1col * 0

                                                                               
        # body conditions                                                      
        self.frames.body                                                       = Conditions()        
        self.frames.body.inertial_rotations                                    = ones_3col * 0
        self.frames.body.thrust_force_vector                                   = ones_3col * 0
        self.frames.body.moment_vector                                         = ones_3col * 0
        self.frames.body.velocity_vector                                       = ones_3col * 0
        self.frames.body.thrust_moment_vector                                  = ones_3col * 0 
        self.frames.body.transform_to_inertial                                 = np.empty([0,0,0])
                                                                               
        # wind frame conditions                                                
        self.frames.wind                                                       = Conditions()
        self.frames.wind.body_rotations                                        = ones_3col * 0 # rotations in [X,Y,Z] -> [phi,theta,psi]
        self.frames.wind.velocity_vector                                       = ones_3col * 0
        self.frames.wind.force_vector                                          = ones_3col * 0
        self.frames.wind.moment_vector                                         = ones_3col * 0
        self.frames.wind.transform_to_inertial                                 = np.empty([0,0,0]) 
                                                                               
        # planet frame conditions                                              
        self.frames.planet                                                     = Conditions()
        self.frames.planet.start_time                                          = None
        self.frames.planet.latitude                                            = ones_1col * 0
        self.frames.planet.longitude                                           = ones_1col * 0
        self.frames.planet.true_course                                         = ones_1col * 0
        self.frames.planet.true_heading                                        = ones_1col * 0

        # ----------------------------------------------------------------------------------------------------------------------         
        # Freestream 
        # ----------------------------------------------------------------------------------------------------------------------    
        self.freestream                                                        = Conditions()        
        self.freestream.velocity                                               = ones_1col * 0
        self.freestream.u                                                      = ones_1col * 0
        self.freestream.v                                                      = ones_1col * 0
        self.freestream.w                                                      = ones_1col * 0
        self.freestream.mach_number                                            = ones_1col * 0
        self.freestream.pressure                                               = ones_1col * 0
        self.freestream.temperature                                            = ones_1col * 0
        self.freestream.density                                                = ones_1col * 0
        self.freestream.speed_of_sound                                         = ones_1col * 0
        self.freestream.dynamic_viscosity                                      = ones_1col * 0
        self.freestream.altitude                                               = ones_1col * 0
        self.freestream.gravity                                                = ones_1col * 0
        self.freestream.reynolds_number                                        = ones_1col * 0
        self.freestream.dynamic_pressure                                       = ones_1col * 0
        self.freestream.delta_ISA                                              = ones_1col * 0

        # ----------------------------------------------------------------------------------------------------------------------         
        # Aerodynamics
        # ----------------------------------------------------------------------------------------------------------------------  
        self.aerodynamics                                                      = Conditions() 
                                                                               
        # aerdynamic angles                                                    
        self.aerodynamics.angles                                               = Conditions()
        self.aerodynamics.angles.alpha                                         = ones_1col * 0
        self.aerodynamics.angles.beta                                          = ones_1col * 0
        self.aerodynamics.angles.phi                                           = ones_1col * 0 
                                                                               
        # aerodynamic coefficients                                             
        self.aerodynamics.coefficients                                         = Conditions()
        self.aerodynamics.coefficients.lift                                    = ones_1col * 0
        self.aerodynamics.coefficients.drag                                    = ones_1col * 0     
                                                                               
        # aerodynamic coefficients                                             
        self.aerodynamics.coefficients                                         = Conditions()
        self.aerodynamics.coefficients.surface_pressure                        = None
        self.aerodynamics.coefficients.lift                                    = Conditions()
        self.aerodynamics.coefficients.lift.total                              = None
        self.aerodynamics.coefficients.lift.induced                            = Conditions()
        self.aerodynamics.coefficients.lift.induced.inviscid_wings             = Conditions()
        self.aerodynamics.coefficients.lift.compressible_wings                 = Conditions() 
        self.aerodynamics.coefficients.drag                                    = Conditions()  
        self.aerodynamics.coefficients.drag.total                              = ones_1col * 0   
        self.aerodynamics.coefficients.drag.parasite                           = Conditions()
        self.aerodynamics.coefficients.drag.compressible                       = Conditions()
        self.aerodynamics.coefficients.drag.induced                            = Conditions()
        self.aerodynamics.coefficients.drag.induced.total                      = ones_1col * 0 
        self.aerodynamics.coefficients.drag.induced.inviscid                   = ones_1col * 0 
        self.aerodynamics.coefficients.drag.induced.inviscid_wings             = Conditions()
        self.aerodynamics.coefficients.drag.cooling                            = Conditions()
        self.aerodynamics.coefficients.drag.cooling.total                      = ones_1col * 0
        self.aerodynamics.coefficients.drag.windmilling                        = Conditions()
        self.aerodynamics.coefficients.drag.windmilling.total                  = ones_1col * 0
        self.aerodynamics.coefficients.drag.asymmetry_trim                     = Conditions()
        self.aerodynamics.coefficients.drag.asymmetry_trim.total               = ones_1col * 0 
        
        self.aerodynamics.coefficients.drag.induced.efficiency_factor          = ones_1col * 0 
        self.aerodynamics.oswald_efficiency                                    = ones_1col * 0 
 
        # ----------------------------------------------------------------------------------------------------------------------
        # Control Surfaces 
        # ----------------------------------------------------------------------------------------------------------------------
        self.control_surfaces                                                  = Conditions()

        self.control_surfaces.aileron                                          = Conditions()
        self.control_surfaces.aileron.deflection                               = ones_1col * 0 
        self.control_surfaces.aileron.static_stability                         = Conditions()
        self.control_surfaces.aileron.static_stability.coefficients            = Conditions() 
        self.control_surfaces.aileron.static_stability.coefficients.lift       = ones_1col * 0        
        self.control_surfaces.aileron.static_stability.coefficients.drag       = ones_1col * 0           
        self.control_surfaces.aileron.static_stability.coefficients.X          = ones_1col * 0           
        self.control_surfaces.aileron.static_stability.coefficients.Y          = ones_1col * 0           
        self.control_surfaces.aileron.static_stability.coefficients.Z          = ones_1col * 0         
        self.control_surfaces.aileron.static_stability.coefficients.L          = ones_1col * 0         
        self.control_surfaces.aileron.static_stability.coefficients.M          = ones_1col * 0         
        self.control_surfaces.aileron.static_stability.coefficients.N          = ones_1col * 0           
        self.control_surfaces.aileron.static_stability.coefficients.e          = ones_1col * 0
        
        self.control_surfaces.elevator                                         = Conditions()
        self.control_surfaces.elevator.deflection                              = ones_1col * 0 
        self.control_surfaces.elevator.static_stability                        = Conditions()
        self.control_surfaces.elevator.static_stability.coefficients           = Conditions() 
        self.control_surfaces.elevator.static_stability.coefficients.lift      = ones_1col * 0        
        self.control_surfaces.elevator.static_stability.coefficients.drag      = ones_1col * 0           
        self.control_surfaces.elevator.static_stability.coefficients.X         = ones_1col * 0           
        self.control_surfaces.elevator.static_stability.coefficients.Y         = ones_1col * 0           
        self.control_surfaces.elevator.static_stability.coefficients.Z         = ones_1col * 0         
        self.control_surfaces.elevator.static_stability.coefficients.L         = ones_1col * 0         
        self.control_surfaces.elevator.static_stability.coefficients.M         = ones_1col * 0         
        self.control_surfaces.elevator.static_stability.coefficients.N         = ones_1col * 0           
        self.control_surfaces.elevator.static_stability.coefficients.e         = ones_1col * 0
        
        self.control_surfaces.rudder                                           = Conditions()
        self.control_surfaces.rudder.deflection                                = ones_1col * 0 
        self.control_surfaces.rudder.static_stability                          = Conditions()
        self.control_surfaces.rudder.static_stability.coefficients             = Conditions() 
        self.control_surfaces.rudder.static_stability.coefficients.lift        = ones_1col * 0         
        self.control_surfaces.rudder.static_stability.coefficients.drag        = ones_1col * 0          
        self.control_surfaces.rudder.static_stability.coefficients.X           = ones_1col * 0          
        self.control_surfaces.rudder.static_stability.coefficients.Y           = ones_1col * 0          
        self.control_surfaces.rudder.static_stability.coefficients.Z           = ones_1col * 0         
        self.control_surfaces.rudder.static_stability.coefficients.L           = ones_1col * 0         
        self.control_surfaces.rudder.static_stability.coefficients.M           = ones_1col * 0         
        self.control_surfaces.rudder.static_stability.coefficients.N           = ones_1col * 0           
        self.control_surfaces.rudder.static_stability.coefficients.e           = ones_1col * 0
        
        self.control_surfaces.flap                                             = Conditions()
        self.control_surfaces.flap.deflection                                  = ones_1col * 0 
        self.control_surfaces.flap.static_stability                            = Conditions()
        self.control_surfaces.flap.static_stability.coefficients               = Conditions() 
        self.control_surfaces.flap.static_stability.coefficients.lift          = ones_1col * 0        
        self.control_surfaces.flap.static_stability.coefficients.drag          = ones_1col * 0           
        self.control_surfaces.flap.static_stability.coefficients.X             = ones_1col * 0           
        self.control_surfaces.flap.static_stability.coefficients.Y             = ones_1col * 0           
        self.control_surfaces.flap.static_stability.coefficients.Z             = ones_1col * 0         
        self.control_surfaces.flap.static_stability.coefficients.L             = ones_1col * 0         
        self.control_surfaces.flap.static_stability.coefficients.M             = ones_1col * 0         
        self.control_surfaces.flap.static_stability.coefficients.N             = ones_1col * 0           
        self.control_surfaces.flap.static_stability.coefficients.e             = ones_1col * 0
        
        self.control_surfaces.slat                                             = Conditions()
        self.control_surfaces.slat.deflection                                  = ones_1col * 0 
        self.control_surfaces.slat.static_stability                            = Conditions()
        self.control_surfaces.slat.static_stability.coefficients               = Conditions() 
        self.control_surfaces.slat.static_stability.coefficients.lift          = ones_1col * 0         
        self.control_surfaces.slat.static_stability.coefficients.drag          = ones_1col * 0          
        self.control_surfaces.slat.static_stability.coefficients.X             = ones_1col * 0          
        self.control_surfaces.slat.static_stability.coefficients.Y             = ones_1col * 0          
        self.control_surfaces.slat.static_stability.coefficients.Z             = ones_1col * 0         
        self.control_surfaces.slat.static_stability.coefficients.L             = ones_1col * 0         
        self.control_surfaces.slat.static_stability.coefficients.M             = ones_1col * 0         
        self.control_surfaces.slat.static_stability.coefficients.N             = ones_1col * 0           
        self.control_surfaces.slat.static_stability.coefficients.e             = ones_1col * 0

        # ----------------------------------------------------------------------------------------------------------------------
        # Stability 
        # ----------------------------------------------------------------------------------------------------------------------  
        self.stability                                                         = Conditions()
        self.static_stability                                                  = Conditions()
 
        self.static_stability.forces                                           = Conditions()
        self.static_stability.forces.lift                                      = ones_1col * 0
        self.static_stability.forces.drag                                      = ones_1col * 0
        self.static_stability.forces.X                                         = ones_1col * 0
        self.static_stability.forces.Y                                         = ones_1col * 0
        self.static_stability.forces.Z                                         = ones_1col * 0
                                                                               
        self.static_stability.moments                                          = Conditions()
        self.static_stability.moments.L                                        = ones_1col * 0
        self.static_stability.moments.M                                        = ones_1col * 0
        self.static_stability.moments.N                                        = ones_1col * 0
                                                                               
        self.static_stability.static_margin                                    = ones_1col * 0
        self.static_stability.neutral_point                                    = ones_1col * 0
        self.static_stability.spiral_criteria                                  = ones_1col * 0 
        self.static_stability.pitch_rate                                       = ones_1col * 0
        self.static_stability.roll_rate                                        = ones_1col * 0
        self.static_stability.yaw_rate                                         = ones_1col * 0 
                                                                               
        self.static_stability.coefficients                                     = Conditions()
        self.static_stability.coefficients.lift                                = ones_1col * 0
        self.static_stability.coefficients.drag                                = ones_1col * 0
        self.static_stability.coefficients.X                                   = ones_1col * 0
        self.static_stability.coefficients.Y                                   = ones_1col * 0
        self.static_stability.coefficients.Z                                   = ones_1col * 0
        self.static_stability.coefficients.L                                   = ones_1col * 0
        self.static_stability.coefficients.M                                   = ones_1col * 0
        self.static_stability.coefficients.N                                   = ones_1col * 0 
        self.static_stability.coefficients.roll                                = ones_1col * 0
        self.static_stability.coefficients.pitch                               = ones_1col * 0
        self.static_stability.coefficients.yaw                                 = ones_1col * 0  
                                                                               
        self.static_stability.derivatives                                      = Conditions()
                                                                               
        # stability axis                                                       
        self.static_stability.derivatives.Clift_alpha                          = ones_1col * 0
        self.static_stability.derivatives.Clift_beta                           = ones_1col * 0
        self.static_stability.derivatives.Clift_delta_a                        = ones_1col * 0
        self.static_stability.derivatives.Clift_delta_e                        = ones_1col * 0
        self.static_stability.derivatives.Clift_delta_r                        = ones_1col * 0
        self.static_stability.derivatives.Clift_delta_f                        = ones_1col * 0
        self.static_stability.derivatives.Clift_delta_s                        = ones_1col * 0
        self.static_stability.derivatives.Cdrag_alpha                          = ones_1col * 0
        self.static_stability.derivatives.Cdrag_beta                           = ones_1col * 0
        self.static_stability.derivatives.Cdrag_delta_a                        = ones_1col * 0
        self.static_stability.derivatives.Cdrag_delta_e                        = ones_1col * 0
        self.static_stability.derivatives.Cdrag_delta_r                        = ones_1col * 0
        self.static_stability.derivatives.Cdrag_delta_f                        = ones_1col * 0
        self.static_stability.derivatives.Cdrag_delta_s                        = ones_1col * 0
        self.static_stability.derivatives.CX_alpha                             = ones_1col * 0
        self.static_stability.derivatives.CX_beta                              = ones_1col * 0
        self.static_stability.derivatives.CX_delta_a                           = ones_1col * 0
        self.static_stability.derivatives.CX_delta_e                           = ones_1col * 0
        self.static_stability.derivatives.CX_delta_r                           = ones_1col * 0
        self.static_stability.derivatives.CX_delta_f                           = ones_1col * 0
        self.static_stability.derivatives.CX_delta_s                           = ones_1col * 0
        self.static_stability.derivatives.CY_alpha                             = ones_1col * 0
        self.static_stability.derivatives.CY_beta                              = ones_1col * 0
        self.static_stability.derivatives.CY_delta_a                           = ones_1col * 0
        self.static_stability.derivatives.CY_delta_e                           = ones_1col * 0
        self.static_stability.derivatives.CY_delta_r                           = ones_1col * 0
        self.static_stability.derivatives.CY_delta_f                           = ones_1col * 0
        self.static_stability.derivatives.CY_delta_s                           = ones_1col * 0
        self.static_stability.derivatives.CZ_alpha                             = ones_1col * 0
        self.static_stability.derivatives.CZ_beta                              = ones_1col * 0
        self.static_stability.derivatives.CZ_delta_a                           = ones_1col * 0
        self.static_stability.derivatives.CZ_delta_e                           = ones_1col * 0
        self.static_stability.derivatives.CZ_delta_r                           = ones_1col * 0
        self.static_stability.derivatives.CZ_delta_f                           = ones_1col * 0
        self.static_stability.derivatives.CZ_delta_s                           = ones_1col * 0
        self.static_stability.derivatives.CL_alpha                             = ones_1col * 0
        self.static_stability.derivatives.CL_beta                              = ones_1col * 0
        self.static_stability.derivatives.CL_delta_a                           = ones_1col * 0
        self.static_stability.derivatives.CL_delta_e                           = ones_1col * 0
        self.static_stability.derivatives.CL_delta_r                           = ones_1col * 0
        self.static_stability.derivatives.CL_delta_f                           = ones_1col * 0
        self.static_stability.derivatives.CL_delta_s                           = ones_1col * 0
        self.static_stability.derivatives.CM_alpha                             = ones_1col * 0
        self.static_stability.derivatives.CM_beta                              = ones_1col * 0
        self.static_stability.derivatives.CM_delta_a                           = ones_1col * 0
        self.static_stability.derivatives.CM_delta_e                           = ones_1col * 0
        self.static_stability.derivatives.CM_delta_r                           = ones_1col * 0
        self.static_stability.derivatives.CM_delta_f                           = ones_1col * 0
        self.static_stability.derivatives.CM_delta_s                           = ones_1col * 0
        self.static_stability.derivatives.CN_alpha                             = ones_1col * 0
        self.static_stability.derivatives.CN_beta                              = ones_1col * 0
        self.static_stability.derivatives.CN_delta_a                           = ones_1col * 0
        self.static_stability.derivatives.CN_delta_e                           = ones_1col * 0
        self.static_stability.derivatives.CN_delta_r                           = ones_1col * 0
        self.static_stability.derivatives.CN_delta_f                           = ones_1col * 0
        self.static_stability.derivatives.CN_delta_s                           = ones_1col * 0
        
        # body axis derivatives
        self.static_stability.derivatives.Clift_u                              = ones_1col * 0
        self.static_stability.derivatives.Clift_v                              = ones_1col * 0
        self.static_stability.derivatives.Clift_w                              = ones_1col * 0 
        self.static_stability.derivatives.Cdrag_u                              = ones_1col * 0
        self.static_stability.derivatives.Cdrag_v                              = ones_1col * 0
        self.static_stability.derivatives.Cdrag_w                              = ones_1col * 0         
        self.static_stability.derivatives.CX_u                                 = ones_1col * 0
        self.static_stability.derivatives.CX_v                                 = ones_1col * 0
        self.static_stability.derivatives.CX_w                                 = ones_1col * 0
        self.static_stability.derivatives.CY_u                                 = ones_1col * 0
        self.static_stability.derivatives.CY_v                                 = ones_1col * 0
        self.static_stability.derivatives.CY_w                                 = ones_1col * 0
        self.static_stability.derivatives.CZ_u                                 = ones_1col * 0
        self.static_stability.derivatives.CZ_v                                 = ones_1col * 0
        self.static_stability.derivatives.CZ_w                                 = ones_1col * 0
        self.static_stability.derivatives.CL_u                                 = ones_1col * 0
        self.static_stability.derivatives.CL_v                                 = ones_1col * 0
        self.static_stability.derivatives.CL_w                                 = ones_1col * 0
        self.static_stability.derivatives.CM_u                                 = ones_1col * 0
        self.static_stability.derivatives.CM_v                                 = ones_1col * 0
        self.static_stability.derivatives.CM_w                                 = ones_1col * 0
        self.static_stability.derivatives.CN_u                                 = ones_1col * 0
        self.static_stability.derivatives.CN_v                                 = ones_1col * 0
        self.static_stability.derivatives.CN_w                                 = ones_1col * 0 
        self.static_stability.derivatives.CZ_alpha_dot                         = ones_1col * 0
        self.static_stability.derivatives.CM_alpha_dot                         = ones_1col * 0 
        self.static_stability.derivatives.Clift_p                              = ones_1col * 0
        self.static_stability.derivatives.Clift_q                              = ones_1col * 0
        self.static_stability.derivatives.Clift_r                              = ones_1col * 0
        self.static_stability.derivatives.Cdrag_p                              = ones_1col * 0
        self.static_stability.derivatives.Cdrag_q                              = ones_1col * 0
        self.static_stability.derivatives.Cdrag_r                              = ones_1col * 0         
        self.static_stability.derivatives.CX_p                                 = ones_1col * 0
        self.static_stability.derivatives.CX_q                                 = ones_1col * 0
        self.static_stability.derivatives.CX_r                                 = ones_1col * 0
        self.static_stability.derivatives.CY_p                                 = ones_1col * 0
        self.static_stability.derivatives.CY_q                                 = ones_1col * 0
        self.static_stability.derivatives.CY_r                                 = ones_1col * 0
        self.static_stability.derivatives.CZ_p                                 = ones_1col * 0
        self.static_stability.derivatives.CZ_q                                 = ones_1col * 0
        self.static_stability.derivatives.CZ_r                                 = ones_1col * 0
        self.static_stability.derivatives.CL_p                                 = ones_1col * 0
        self.static_stability.derivatives.CL_q                                 = ones_1col * 0
        self.static_stability.derivatives.CL_r                                 = ones_1col * 0
        self.static_stability.derivatives.CM_p                                 = ones_1col * 0
        self.static_stability.derivatives.CM_q                                 = ones_1col * 0
        self.static_stability.derivatives.CM_r                                 = ones_1col * 0
        self.static_stability.derivatives.CN_p                                 = ones_1col * 0
        self.static_stability.derivatives.CN_q                                 = ones_1col * 0
        self.static_stability.derivatives.CN_r                                 = ones_1col * 0 
                                                                               
        # dynamic stability                                                    
        self.dynamic_stability                                                 = Conditions()
        self.dynamic_stability.LongModes                                       = Conditions()
        self.dynamic_stability.LatModes                                        = Conditions()
        

        # ----------------------------------------------------------------------------------------------------------------------         
        # Freestream 
        # ----------------------------------------------------------------------------------------------------------------------    
        self.emissions                                       = Conditions()         
        
        # ----------------------------------------------------------------------------------------------------------------------         
        # Noise
        # ----------------------------------------------------------------------------------------------------------------------       
        self.noise                                            = Conditions() 

        # ----------------------------------------------------------------------------------------------------------------------         
        # Energy
        # ---------------------------------------------------------------------------------------------------------------------- 
        self.energy                                           = Conditions()
        self.energy.throttle                                  = ones_1col * 0  
        self.energy.thrust_breakdown                          = Conditions()
        self.energy.thrust_breakdown                          = Conditions()
        self.energy.thrust_force_vector                       = ones_3col * 0
        self.energy.thrust_moment_vector                      = ones_3col * 0
        self.energy.power                                     = ones_1col * 0
        self.energy.vehicle_mass_rate                         = ones_1col * 0
        
        
        # ----------------------------------------------------------------------------------------------------------------------         
        # Weights 
        # ----------------------------------------------------------------------------------------------------------------------     
        self.weights                                          = Conditions() 
        self.weights.total_mass                               = ones_1col * 0
        self.weights.total_moment_of_inertia                  = ones_3col * 0 # 3 total I(I_xx, I_yy, I_zz)? or 9(including I_xz etc)?
        self.weights.weight_breakdown                         = Conditions()
        self.weights.vehicle_mass_rate                        = ones_1col * 0


# RCAIDE/Compoments/Landing_Gear/Nose_Landing_Gear.py
# 
# Created:  Nov 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports     
from .Landing_Gear import Landing_Gear

# ---------------------------------------------------------------------------------------------------------------------- 
#  Nose_Landing_Gear
# ----------------------------------------------------------------------------------------------------------------------   
class Nose_Landing_Gear(Landing_Gear):
    """
    Nose landing gear component that provides directional control during ground operations 
    and supports a portion of the aircraft's forward weight.

    Attributes
    ----------
    tag : str
        Unique identifier for the nose landing gear component, defaults to 'nose_gear'
        
    tire_diameter : float
        Diameter of the nose gear tires, defaults to 0
        
    strut_length : float
        Length of the nose gear strut assembly, defaults to 0
        
    units : float
        Number of nose landing gear units, typically 1, defaults to 0
        
    wheels : float
        Number of wheels per nose gear unit, defaults to 0

    Notes
    -----
    The nose landing gear typically carries 10-20% of the aircraft weight and provides 
    steering capability during taxi, takeoff, and landing.
    
    **Major Assumptions**
    
    * Located on aircraft centerline
    * Provides steering capability
    * Designed for lower loads than main gear
    
    **Definitions**
        
    'Steering Angle'
        Maximum deflection angle available for directional control

    See Also
    --------
    RCAIDE.Library.Components.Landing_Gear.Landing_Gear
        Base landing gear class
    RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear
        Main gear component that works in conjunction with nose gear
    """

    def __defaults__(self): 
        """
        Sets default values for the nose landing gear attributes.
        """
        self.tag           = 'nose_gear'
        self.tire_diameter = 0.    
        self.strut_length  = 0.    
        self.units         = 0. # number of nose landing gear    
        self.wheels        = 0. # number of wheels on the nose landing gear 


# RCAIDE/Methods/Aerodynamics/Airfoil_Panel_Method/airfoil_analysis.py
# 
# 
# Created:  Dec 2023, M. Clarke
# Modified: Apr 2023, N. Nanjappa

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Core import Data
from .hess_smith           import hess_smith
from .thwaites_method      import thwaites_method 
from .heads_method         import heads_method 
from .aero_coeff           import aero_coeff
from .chordwise_distribution import chordwise_distribution
from .cf_filter            import cf_filter

# pacakge imports  
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
# airfoil_analysis.py
# ----------------------------------------------------------------------------------------------------------------------
def airfoil_analysis(airfoil_geometry,alpha,Re_L,initial_momentum_thickness=1E-5,tolerance = 1E0,H_wake = 1.05,Ue_wake = 0.99):
    """This computes the aerodynamic polars as well as the boundary layer properties of 
    an airfoil at a defined set of reynolds numbers and angle of attacks

    Assumptions:
    Michel Criteria used for transition 
    
    Squire-Young relation for total drag (exrapolates theta from end of wake). 
    However, since we do not have a wake we will assume H_wake = 1.05 and Ue_wake = 0.99
    The freestream velocity is taken to be 1 m/s.
    Airfoil has a unit chord length [1 m].

    Source:
    N/A

    Inputs: 
    airfoil_geometry   - airfoil geometry points                                                             [unitless]
    alpha              - angle of attacks                                                                    [radians]
    Re_L               - Reynolds numbers                                                                    [unitless]
    batch_analysis     - boolean : If True: the specified number of angle of attacks and Reynolds            [boolean]
                                  numbers are used to create a table of 2-D results for each combination
                                  Note: Can only accomodate one airfoil
                                  
                                  If False:The airfoils specified are run and corresponding angle of attacks 
                                  and Reynolds numbers
                                  Note: The number of airfoils, angle of attacks and reynolds numbers must 
                                  all the same dimension                     
    
    Outputs: 
    airfoil_properties.
        AoA            - angle of attack                                                   [radians
        Re             - Reynolds number                                                   [unitless]
        Cl             - lift coefficients                                                 [unitless]
        Cd             - drag coefficients                                                 [unitless]
        Cm             - moment coefficients                                               [unitless]
        normals        - surface normals of airfoil                                        [unitless]
        x              - x coordinate points on airfoil                                    [unitless]
        y              - y coordinate points on airfoil                                    [unitless]
        x_bl           - x coordinate points on airfoil adjusted to include boundary layer [unitless]
        y_bl           - y coordinate points on airfoil adjusted to include boundary layer [unitless]
        Cp             - pressure coefficient distribution                                 [unitless]
        Ue_Vinf        - ratio of boundary layer edge velocity to freestream               [unitless]
        dVe            - derivative of boundary layer velocity                             [m/s-m]
        theta          - momentum thickness                                                [m]
        delta_star     - displacement thickness                                            [m]
        delta          - boundary layer thickness                                          [m]
        H              - shape factor                                                      [unitless]
        Cf             - local skin friction coefficient                                   [unitless]
        Re_theta_t     - Reynolds Number as a function of theta transition location        [unitless]
        tr_crit        - critical transition criteria                                      [unitless]
                        
    Properties Used:
    N/A
    """     
    ncases       = len(alpha[0,:]) 
    ncpts        = len(Re_L)  
    x_coord      = airfoil_geometry.x_coordinates
    y_coord      = airfoil_geometry.y_coordinates
    npanel       = len(x_coord)-1 
               
    x_coord_3d = np.tile(x_coord[:,None,None],(1,ncases,ncpts)) # number of points, number of cases, number of control points 
    y_coord_3d = np.tile(y_coord[:,None,None],(1,ncases,ncpts)) # number of points, number of cases, number of control points 
        
    # Begin by solving for velocity distribution at airfoil surface using inviscid panel simulation
    # these are the locations (faces) where things are computed , len = n panel
    # dimension of vt = npanel x ncases x ncpts
    X,Y,vt,normals = hess_smith(x_coord_3d,y_coord_3d,alpha,Re_L,npanel)  
    
    # Reynolds number 
    RE_L_VALS = Re_L.T 
    
    # kinematic coefficient of viscosity (assuming unit chord and unit inlet velocity)
    nu           = 1/RE_L_VALS
    
    # ---------------------------------------------------------------------
    # Bottom surface of airfoil 
    # ---------------------------------------------------------------------     
    VT              = np.ma.masked_greater(vt,0 )
    VT_mask         = np.ma.masked_greater(vt,0 ).mask
    X_BOT_VALS      = np.ma.array(X, mask = VT_mask)[::-1]
    Y_BOT           = np.ma.array(Y, mask = VT_mask)[::-1]
         
    X_BOT           = np.zeros_like(X_BOT_VALS)
    X_BOT[1:]       = np.cumsum(np.sqrt((X_BOT_VALS[1:] - X_BOT_VALS[:-1])**2 + (Y_BOT[1:] - Y_BOT[:-1])**2),axis = 0)
    first_idx       = np.ma.count_masked(X_BOT,axis = 0)
    mask_count      = np.ma.count(X_BOT,axis = 0)
    prev_index      = first_idx-1
    first_panel     = list(prev_index.flatten())
    last_panel      = list((first_idx-1 + mask_count).flatten())
    last_paneldve   = list((first_idx-2 + mask_count).flatten())
    aoas            = list(np.repeat(np.arange(ncases),ncpts))
    res             = list(np.tile(np.arange(ncpts),ncases) )
    X_BOT.mask[first_panel,aoas,res] = False
    
    # flow velocity and pressure of on botton surface 
    VE_BOT          = -VT[::-1]  
    
    # velocity gradients on bottom surface  
    DVE_BOT                        = np.ma.zeros(np.shape(X_BOT)) 
    DVE_BOT_TEMP                   = np.diff(VE_BOT,axis = 0)/np.diff(X_BOT,axis = 0)
    a                              = X_BOT[1:-1] - X_BOT[:-2]
    b                              = X_BOT[2:]   - X_BOT[:-2]
    DVE_BOT[1:-1]                  = ((b*DVE_BOT_TEMP[:-1] + a*DVE_BOT_TEMP[1:])/(a+b)).data
    DVE_BOT.mask                   = X_BOT.mask
    DVE_BOT[first_panel,aoas,res]  = DVE_BOT_TEMP[first_panel,aoas,res]
    DVE_BOT[last_panel,aoas,res]   = DVE_BOT_TEMP[last_paneldve,aoas,res] 
    
    # x - location of stagnation point 
    L_BOT                          = X_BOT[-1,:,:]    
        
    # laminar boundary layer properties using thwaites method  
    BOT_T_RESULTS  = thwaites_method(npanel,ncases,ncpts, nu, L_BOT , RE_L_VALS, X_BOT, VE_BOT, DVE_BOT,tolerance,
                                      THETA_0=initial_momentum_thickness)
    
    X_T_BOT          = BOT_T_RESULTS.X_T      
    THETA_T_BOT      = BOT_T_RESULTS.THETA_T     
    DELTA_STAR_T_BOT = BOT_T_RESULTS.DELTA_STAR_T  
    H_T_BOT          = BOT_T_RESULTS.H_T         
    CF_T_BOT         = BOT_T_RESULTS.CF_T   
    RE_THETA_T_BOT   = BOT_T_RESULTS.RE_THETA_T    
    RE_X_T_BOT       = BOT_T_RESULTS.RE_X_T      
    DELTA_T_BOT      = BOT_T_RESULTS.DELTA_T      
     
    # transition location  
    TR_CRIT_BOT      = RE_THETA_T_BOT - 1.174*(1 + 22400/RE_X_T_BOT)*RE_X_T_BOT**0.46  
    CRITERION_BOT    = np.ma.masked_greater(TR_CRIT_BOT,0 )  
    mask_count       = np.ma.count(CRITERION_BOT,axis = 0)  
    mask_count[mask_count == npanel] = npanel-1 
    transition_panel = list(mask_count.flatten()) 
    aoas             = list(np.repeat(np.arange(ncases),ncpts))
    res              = list(np.tile(np.arange(ncpts),ncases))
        
    X_TR_BOT          = X_T_BOT[transition_panel,aoas,res].reshape(ncases,ncpts)
    DELTA_STAR_TR_BOT = DELTA_STAR_T_BOT[transition_panel,aoas,res].reshape(ncases,ncpts)
    THETA_TR_BOT      = THETA_T_BOT[transition_panel,aoas,res].reshape(ncases,ncpts)    
    DELTA_TR_BOT      = DELTA_T_BOT[transition_panel,aoas,res].reshape(ncases,ncpts) 
    CF_TR_BOT         = CF_T_BOT[transition_panel,aoas,res].reshape(ncases,ncpts)
    H_TR_BOT          = H_T_BOT[transition_panel,aoas,res].reshape(ncases,ncpts)
    
    # TURBULENT_SURF  
    TURBULENT_SURF    = L_BOT.data
    TURBULENT_COORD   = np.ma.masked_less(X_BOT.data  - X_TR_BOT,0) 
    
    # turbulent boundary layer properties using heads method  
    BOT_H_RESULTS     = heads_method(npanel,ncases,ncpts, nu, DELTA_TR_BOT, THETA_TR_BOT , DELTA_STAR_TR_BOT, 
                                         CF_TR_BOT, H_TR_BOT, RE_L_VALS, TURBULENT_COORD, VE_BOT, DVE_BOT, TURBULENT_SURF, tolerance)
    
    X_H_BOT          = BOT_H_RESULTS.X_H      
    THETA_H_BOT      = BOT_H_RESULTS.THETA_H   
    DELTA_STAR_H_BOT = BOT_H_RESULTS.DELTA_STAR_H
    H_H_BOT          = BOT_H_RESULTS.H_H       
    CF_H_BOT         = BOT_H_RESULTS.CF_H    
    RE_THETA_H_BOT   = BOT_H_RESULTS.RE_THETA_H 
    RE_X_H_BOT       = BOT_H_RESULTS.RE_X_H         
    DELTA_H_BOT      = BOT_H_RESULTS.DELTA_H   
    
    # Apply Masks to surface vectors  
    X_T_BOT          = np.ma.array(X_T_BOT , mask = CRITERION_BOT.mask)
    THETA_T_BOT      = np.ma.array(THETA_T_BOT , mask = CRITERION_BOT.mask)   
    DELTA_STAR_T_BOT = np.ma.array(DELTA_STAR_T_BOT , mask = CRITERION_BOT.mask)
    H_T_BOT          = np.ma.array(H_T_BOT , mask = CRITERION_BOT.mask)       
    CF_T_BOT         = np.ma.array(CF_T_BOT , mask = CRITERION_BOT.mask)      
    RE_THETA_T_BOT   = np.ma.array(RE_THETA_T_BOT , mask = CRITERION_BOT.mask) 
    RE_X_T_BOT       = np.ma.array(RE_X_T_BOT  , mask = CRITERION_BOT.mask)   
    DELTA_T_BOT      = np.ma.array(DELTA_T_BOT , mask = CRITERION_BOT.mask)    
                       
    X_H_BOT          = np.ma.array(X_H_BOT , mask = ~CRITERION_BOT.mask)       
    THETA_H_BOT      = np.ma.array(THETA_H_BOT , mask = ~CRITERION_BOT.mask)   
    DELTA_STAR_H_BOT = np.ma.array(DELTA_STAR_H_BOT , mask = ~CRITERION_BOT.mask)
    H_H_BOT          = np.ma.array(H_H_BOT  , mask = ~CRITERION_BOT.mask)      
    CF_H_BOT         = np.ma.array(CF_H_BOT , mask = ~CRITERION_BOT.mask)  
    RE_THETA_H_BOT   = np.ma.array(RE_THETA_H_BOT  , mask = ~CRITERION_BOT .mask)    
    RE_X_H_BOT       = np.ma.array(RE_X_H_BOT  , mask = ~CRITERION_BOT .mask)        
    DELTA_H_BOT      = np.ma.array(DELTA_H_BOT , mask = ~CRITERION_BOT.mask)   
        
    
    # Concatenate vectors
    X_H_BOT_MOD = X_H_BOT.data + X_TR_BOT.data
    X_H_BOT_MOD = np.ma.array(X_H_BOT_MOD, mask = X_H_BOT.mask)
    
    X_BOT_SURF           = np.ma.concatenate([X_T_BOT, X_H_BOT_MOD], axis = 0 ) 
    THETA_BOT_SURF       = np.ma.concatenate([THETA_T_BOT,THETA_H_BOT ], axis = 0)
    DELTA_STAR_BOT_SURF  = np.ma.concatenate([DELTA_STAR_T_BOT,DELTA_STAR_H_BOT], axis = 0)
    H_BOT_SURF           = np.ma.concatenate([H_T_BOT,H_H_BOT], axis = 0)
    CF_BOT_SURF          = np.ma.concatenate([CF_T_BOT,CF_H_BOT], axis = 0)
    RE_THETA_BOT_SURF    = np.ma.concatenate([RE_THETA_T_BOT,RE_THETA_H_BOT], axis = 0)
    RE_X_BOT_SURF        = np.ma.concatenate([RE_X_T_BOT,RE_X_H_BOT], axis = 0)    
    DELTA_BOT_SURF       = np.ma.concatenate([DELTA_T_BOT,DELTA_H_BOT], axis = 0) 
    
    # Flatten array to extract values 
    X_BOT_SURF_1          = X_BOT_SURF.flatten('F')
    THETA_BOT_SURF_1      = THETA_BOT_SURF.flatten('F') 
    DELTA_STAR_BOT_SURF_1 = DELTA_STAR_BOT_SURF.flatten('F') 
    H_BOT_SURF_1          = H_BOT_SURF.flatten('F')
    CF_BOT_SURF_1         = CF_BOT_SURF.flatten('F')
    RE_THETA_BOT_SURF_1   = RE_THETA_BOT_SURF.flatten('F') 
    RE_X_BOT_SURF_1       = RE_X_BOT_SURF.flatten('F')
    DELTA_BOT_SURF_1      = DELTA_BOT_SURF.flatten('F')  
    
    # Extract values that are not masked 
    X_BOT_SURF_2           = X_BOT_SURF_1.data[~X_BOT_SURF_1.mask]
    THETA_BOT_SURF_2       = THETA_BOT_SURF_1.data[~THETA_BOT_SURF_1.mask] 
    DELTA_STAR_BOT_SURF_2  = DELTA_STAR_BOT_SURF_1.data[~DELTA_STAR_BOT_SURF_1.mask]
    H_BOT_SURF_2           = H_BOT_SURF_1.data[~H_BOT_SURF_1.mask]
    CF_BOT_SURF_2          = CF_BOT_SURF_1.data[~CF_BOT_SURF_1.mask]
    RE_THETA_BOT_SURF_2    = RE_THETA_BOT_SURF_1.data[~RE_THETA_BOT_SURF_1.mask] 
    RE_X_BOT_SURF_2        = RE_X_BOT_SURF_1.data[~RE_X_BOT_SURF_1.mask]
    DELTA_BOT_SURF_2       = DELTA_BOT_SURF_1.data[~DELTA_BOT_SURF_1.mask]
    
    X_BOT_SURF           = X_BOT_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')   
    THETA_BOT_SURF       = THETA_BOT_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')  
    DELTA_STAR_BOT_SURF  = DELTA_STAR_BOT_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')  
    H_BOT_SURF           = H_BOT_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')
    CF_BOT_SURF          = CF_BOT_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')
    RE_THETA_BOT_SURF    = RE_THETA_BOT_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')  
    RE_X_BOT_SURF        = RE_X_BOT_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')  
    DELTA_BOT_SURF       = DELTA_BOT_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')          
    
    
    # ------------------------------------------------------------------------------------------------------
    # Top surface of airfoil 
    # ------------------------------------------------------------------------------------------------------ 
    VT              = np.ma.masked_less(vt,0 )
    CP              = 1 - vt**2 
    VT_mask         = np.ma.masked_less(vt,0 ).mask
    X_TOP_VALS      = np.ma.array(X, mask = VT_mask) 
    Y_TOP           = np.ma.array(Y, mask = VT_mask)  

    X_TOP           = np.zeros_like(X_TOP_VALS)
    X_TOP[1:]       = np.cumsum(np.sqrt((X_TOP_VALS[1:] - X_TOP_VALS[:-1])**2 + (Y_TOP[1:] - Y_TOP[:-1])**2),axis = 0)
    first_idx       = np.ma.count_masked(X_TOP,axis = 0)
    mask_count      = np.ma.count(X_TOP,axis = 0)
    prev_index      = first_idx-1
    first_panel     = list(prev_index.flatten())
    last_panel      = list((first_idx-1 + mask_count).flatten())
    last_paneldve   = list((first_idx-2 + mask_count).flatten())
    aoas            = list(np.repeat(np.arange(ncases),ncpts))
    res             = list(np.tile(np.arange(ncpts),ncases) )
    X_TOP.mask[first_panel,aoas,res] = False
    
    # flow velocity and pressure of on botton surface 
    VE_TOP          = VT    
    
    # velocity gradients on bottom surface  
    DVE_TOP                        = np.ma.zeros(np.shape(X_TOP)) 
    DVE_TOP_TEMP                   = np.diff(VE_TOP,axis = 0)/np.diff(X_TOP,axis = 0)
    a                              = X_TOP[1:-1] - X_TOP[:-2]
    b                              = X_TOP[2:]   - X_TOP[:-2]
    DVE_TOP[1:-1]                  = ((b*DVE_TOP_TEMP[:-1] + a*DVE_TOP_TEMP[1:])/(a+b)).data
    DVE_TOP.mask                   = X_TOP.mask
    DVE_TOP[first_panel,aoas,res]  = DVE_TOP_TEMP[first_panel,aoas,res]
    DVE_TOP[last_panel,aoas,res]   = DVE_TOP_TEMP[last_paneldve,aoas,res] 

    # x - location of stagnation point 
    L_TOP                          = X_TOP[-1,:,:]    

    # laminar boundary layer properties using thwaites method 
    TOP_T_RESULTS    = thwaites_method(npanel,ncases,ncpts, nu, L_TOP , RE_L_VALS,X_TOP,VE_TOP, DVE_TOP,tolerance,
                                      THETA_0=initial_momentum_thickness) 
    
    X_T_TOP          = TOP_T_RESULTS.X_T      
    THETA_T_TOP      = TOP_T_RESULTS.THETA_T     
    DELTA_STAR_T_TOP = TOP_T_RESULTS.DELTA_STAR_T  
    H_T_TOP          = TOP_T_RESULTS.H_T         
    CF_T_TOP         = TOP_T_RESULTS.CF_T      
    RE_THETA_T_TOP   = TOP_T_RESULTS.RE_THETA_T    
    RE_X_T_TOP       = TOP_T_RESULTS.RE_X_T      
    DELTA_T_TOP      = TOP_T_RESULTS.DELTA_T      

    # transition location  
    TR_CRIT_TOP       = RE_THETA_T_TOP - 1.174*(1 + 22400/RE_X_T_TOP)*(RE_X_T_TOP**0.46)  
    CRITERION_TOP     = np.ma.masked_greater(TR_CRIT_TOP,0 ) 
    mask_count        = np.ma.count(CRITERION_TOP,axis = 0)  
    mask_count[mask_count == npanel] = npanel-1  
    transition_panel  = list(mask_count.flatten()) 
    aoas              = list(np.repeat(np.arange(ncases),ncpts))
    res               = list(np.tile(np.arange(ncpts),ncases) )
  
    X_TR_TOP          = X_T_TOP[transition_panel,aoas,res].reshape(ncases,ncpts)
    DELTA_STAR_TR_TOP = DELTA_STAR_T_TOP[transition_panel,aoas,res].reshape(ncases,ncpts)
    THETA_TR_TOP      = THETA_T_TOP[transition_panel,aoas,res].reshape(ncases,ncpts)    
    DELTA_TR_TOP      = DELTA_T_TOP[transition_panel,aoas,res].reshape(ncases,ncpts)
    CF_TR_TOP         = CF_T_TOP[transition_panel,aoas,res].reshape(ncases,ncpts)
    H_TR_TOP          = H_T_TOP[transition_panel,aoas,res].reshape(ncases,ncpts)
   
    TURBULENT_SURF    = L_TOP.data
    TURBULENT_COORD   = np.ma.masked_less( X_TOP.data  - X_TR_TOP,0)

    # turbulent boundary layer properties using heads method  
    TOP_H_RESULTS     = heads_method(npanel,ncases,ncpts, nu, DELTA_TR_TOP, THETA_TR_TOP , DELTA_STAR_TR_TOP, 
                                         CF_TR_TOP, H_TR_TOP, RE_L_VALS, TURBULENT_COORD, VE_TOP, DVE_TOP, TURBULENT_SURF, tolerance)

    X_H_TOP          = TOP_H_RESULTS.X_H      
    THETA_H_TOP      = TOP_H_RESULTS.THETA_H   
    DELTA_STAR_H_TOP = TOP_H_RESULTS.DELTA_STAR_H
    H_H_TOP          = TOP_H_RESULTS.H_H       
    CF_H_TOP         = TOP_H_RESULTS.CF_H        
    RE_THETA_H_TOP   = TOP_H_RESULTS.RE_THETA_H  
    RE_X_H_TOP       = TOP_H_RESULTS.RE_X_H     
    DELTA_H_TOP      = TOP_H_RESULTS.DELTA_H 
    
    # Apply Masks to surface vectors  
    X_T_TOP          = np.ma.array(X_T_TOP , mask = CRITERION_TOP.mask)
    THETA_T_TOP      = np.ma.array(THETA_T_TOP , mask = CRITERION_TOP.mask)   
    DELTA_STAR_T_TOP = np.ma.array(DELTA_STAR_T_TOP , mask = CRITERION_TOP.mask)
    H_T_TOP          = np.ma.array(H_T_TOP , mask = CRITERION_TOP.mask)       
    CF_T_TOP         = np.ma.array(CF_T_TOP , mask = CRITERION_TOP.mask)      
    RE_THETA_T_TOP   = np.ma.array(RE_THETA_T_TOP , mask = CRITERION_TOP.mask) 
    RE_X_T_TOP       = np.ma.array(RE_X_T_TOP  , mask = CRITERION_TOP.mask)   
    DELTA_T_TOP      = np.ma.array(DELTA_T_TOP , mask = CRITERION_TOP.mask)    
                       
    X_H_TOP          = np.ma.array(X_H_TOP , mask = ~CRITERION_TOP.mask)       
    THETA_H_TOP      = np.ma.array(THETA_H_TOP , mask = ~CRITERION_TOP.mask)   
    DELTA_STAR_H_TOP = np.ma.array(DELTA_STAR_H_TOP , mask = ~CRITERION_TOP.mask)
    H_H_TOP          = np.ma.array(H_H_TOP  , mask = ~CRITERION_TOP.mask)      
    CF_H_TOP         = np.ma.array(CF_H_TOP , mask = ~CRITERION_TOP.mask)    
    RE_THETA_H_TOP   = np.ma.array(RE_THETA_H_TOP , mask = ~CRITERION_TOP.mask)    
    RE_X_H_TOP       = np.ma.array(RE_X_H_TOP , mask = ~CRITERION_TOP.mask)    
    DELTA_H_TOP      = np.ma.array(DELTA_H_TOP , mask = ~CRITERION_TOP.mask)   
    
    # Concatenate laminar and turbulent vectors
    X_H_TOP_MOD          = X_H_TOP.data + X_TR_TOP.data
    X_H_TOP_MOD          = np.ma.array(X_H_TOP_MOD, mask = X_H_TOP.mask)
    
    X_TOP_SURF           = np.ma.concatenate([X_T_TOP, X_H_TOP_MOD], axis = 0 ) 
    THETA_TOP_SURF       = np.ma.concatenate([THETA_T_TOP,THETA_H_TOP ], axis = 0)
    DELTA_STAR_TOP_SURF  = np.ma.concatenate([DELTA_STAR_T_TOP,DELTA_STAR_H_TOP], axis = 0)
    H_TOP_SURF           = np.ma.concatenate([H_T_TOP,H_H_TOP], axis = 0)
    CF_TOP_SURF          = np.ma.concatenate([CF_T_TOP,CF_H_TOP], axis = 0)
    RE_THETA_TOP_SURF    = np.ma.concatenate([RE_THETA_T_TOP,RE_THETA_H_TOP], axis = 0)
    RE_X_TOP_SURF        = np.ma.concatenate([RE_X_T_TOP,RE_X_H_TOP], axis = 0)
    DELTA_TOP_SURF       = np.ma.concatenate([DELTA_T_TOP,DELTA_H_TOP], axis = 0) 
    
    # Flatten array to extract values 
    X_TOP_SURF_1          = X_TOP_SURF.flatten('F')
    THETA_TOP_SURF_1      = THETA_TOP_SURF.flatten('F') 
    DELTA_STAR_TOP_SURF_1 = DELTA_STAR_TOP_SURF.flatten('F') 
    H_TOP_SURF_1          = H_TOP_SURF.flatten('F')
    CF_TOP_SURF_1         = CF_TOP_SURF.flatten('F')
    RE_THETA_TOP_SURF_1   = RE_THETA_TOP_SURF.flatten('F') 
    RE_X_TOP_SURF_1       = RE_X_TOP_SURF.flatten('F')
    DELTA_TOP_SURF_1      = DELTA_TOP_SURF.flatten('F')  
    
    # Extract values that are not masked 
    X_TOP_SURF_2           = X_TOP_SURF_1.data[~X_TOP_SURF_1.mask]
    THETA_TOP_SURF_2       = THETA_TOP_SURF_1.data[~THETA_TOP_SURF_1.mask] 
    DELTA_STAR_TOP_SURF_2  = DELTA_STAR_TOP_SURF_1.data[~DELTA_STAR_TOP_SURF_1.mask]
    H_TOP_SURF_2           = H_TOP_SURF_1.data[~H_TOP_SURF_1.mask]
    CF_TOP_SURF_2          = CF_TOP_SURF_1.data[~CF_TOP_SURF_1.mask]
    RE_THETA_TOP_SURF_2    = RE_THETA_TOP_SURF_1.data[~RE_THETA_TOP_SURF_1.mask] 
    RE_X_TOP_SURF_2        = RE_X_TOP_SURF_1.data[~RE_X_TOP_SURF_1.mask]
    DELTA_TOP_SURF_2       = DELTA_TOP_SURF_1.data[~DELTA_TOP_SURF_1.mask]
    
    X_TOP_SURF           = X_TOP_SURF_2.reshape((npanel,ncases,ncpts),order = 'F') 
    THETA_TOP_SURF       = THETA_TOP_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')   
    DELTA_STAR_TOP_SURF  = DELTA_STAR_TOP_SURF_2.reshape((npanel,ncases,ncpts),order = 'F') 
    H_TOP_SURF           = H_TOP_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')
    CF_TOP_SURF          = CF_TOP_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')
    RE_THETA_TOP_SURF    = RE_THETA_TOP_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')  
    RE_X_TOP_SURF        = RE_X_TOP_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')  
    DELTA_TOP_SURF       = DELTA_TOP_SURF_2.reshape((npanel,ncases,ncpts),order = 'F')          

    
    # ------------------------------------------------------------------------------------------------------
    # concatenate lower and upper surfaces   
    # ------------------------------------------------------------------------------------------------------ 
    THETA      = concatenate_surfaces(X_BOT,X_TOP,THETA_BOT_SURF,THETA_TOP_SURF,npanel,ncases,ncpts)
    DELTA_STAR = concatenate_surfaces(X_BOT,X_TOP,DELTA_STAR_BOT_SURF,DELTA_STAR_TOP_SURF,npanel,ncases,ncpts) 
    H          = concatenate_surfaces(X_BOT,X_TOP,H_BOT_SURF,H_TOP_SURF,npanel,ncases,ncpts)  
    CF         = concatenate_surfaces(X_BOT,X_TOP,CF_BOT_SURF,CF_TOP_SURF,npanel,ncases,ncpts) 
    RE_THETA   = concatenate_surfaces(X_BOT,X_TOP,RE_THETA_BOT_SURF,RE_THETA_TOP_SURF,npanel,ncases,ncpts)  
    RE_X       = concatenate_surfaces(X_BOT,X_TOP,RE_X_BOT_SURF,RE_X_TOP_SURF,npanel,ncases,ncpts) 
    DELTA      = concatenate_surfaces(X_BOT,X_TOP,DELTA_BOT_SURF,DELTA_TOP_SURF,npanel,ncases,ncpts)   
     
    VE_VALS    = np.ma.concatenate([np.flip(VE_BOT,axis = 0),VE_TOP ], axis = 0)
    DVE_VALS   = np.ma.concatenate([np.flip(DVE_BOT,axis = 0),DVE_TOP], axis = 0)    
    VE_VALS_1  = VE_VALS.flatten('F')
    DVE_VALS_1 = DVE_VALS.flatten('F')   
    VE_VALS_2  = VE_VALS_1.data[~VE_VALS_1.mask]
    DVE_VALS_2 = DVE_VALS_1.data[~DVE_VALS_1.mask]  
    VE         = VE_VALS_2.reshape((npanel,ncases,ncpts),order = 'F') 
    DVE        = DVE_VALS_2.reshape((npanel,ncases,ncpts),order = 'F')  
    
    
    # ------------------------------------------------------------------------------------------------------
    # Filter Skin Friction Coeffiecient
    # ------------------------------------------------------------------------------------------------------
    CF         = cf_filter(ncpts, ncases, npanel, CF)
    
    
    # ------------------------------------------------------------------------------------------------------
    # Compute effective surface of airfoil with boundary layer and recompute aerodynamic properties  
    # ------------------------------------------------------------------------------------------------------   
    DELTA          = np.nan_to_num(DELTA) # make sure no nans   
    y_coord_3d_bl  = Y + DELTA*normals[:,1,:,:]
    x_coord_3d_bl  = X + DELTA*normals[:,0,:,:]
    npanel_mod     = npanel-1 
    
    X_BL,Y_BL,vt_bl,normals_bl = hess_smith(x_coord_3d_bl,y_coord_3d_bl,alpha,Re_L,npanel_mod)      
      
    # ---------------------------------------------------------------------
    # Bottom surface of airfoil with boundary layer 
    # ---------------------------------------------------------------------       
    VT_BL           = np.ma.masked_greater(vt_bl,0 )
    VT_BL_mask      = np.ma.masked_greater(vt_bl,0 ).mask
    X_BL_BOT_VALS   = np.ma.array(X_BL, mask = VT_BL_mask)[::-1]
    Y_BL_BOT        = np.ma.array(Y_BL, mask = VT_BL_mask)[::-1] 
    X_BL_BOT        = np.zeros_like(X_BL_BOT_VALS)
    X_BL_BOT[1:]    = np.cumsum(np.sqrt((X_BL_BOT_VALS[1:] - X_BL_BOT_VALS[:-1])**2 + (Y_BL_BOT[1:] - Y_BL_BOT[:-1])**2),axis = 0)
    first_idx       = np.ma.count_masked(X_BL_BOT,axis = 0)
    mask_count      = np.ma.count(X_BL_BOT,axis = 0)
    prev_index      = first_idx-1
    first_panel     = list(prev_index.flatten())
    last_panel      = list((first_idx-1 + mask_count).flatten())
    last_paneldve   = list((first_idx-2 + mask_count).flatten())
    aoas            = list(np.repeat(np.arange(ncases),ncpts))
    res             = list(np.tile(np.arange(ncpts),ncases) )
    X_BL_BOT.mask[first_panel,aoas,res] = False
    
    # flow velocity and pressure of on bottom surface 
    VE_BL_BOT          = -VT_BL[::-1] 
    CP_BL_BOT          = 1 - VE_BL_BOT**2   
         
    # ---------------------------------------------------------------------
    # Top surface of airfoil with boundary layer 
    # ---------------------------------------------------------------------    
    VT_BL           = np.ma.masked_less(vt_bl,0 )
    VT_BL_mask      = np.ma.masked_less(vt_bl,0 ).mask
    X_BL_TOP_VALS   = np.ma.array(X_BL, mask = VT_BL_mask) 
    Y_BL_TOP        = np.ma.array(Y_BL, mask = VT_BL_mask)  
 
    X_BL_TOP        = np.zeros_like(X_BL_TOP_VALS)
    X_BL_TOP[1:]    = np.cumsum(np.sqrt((X_BL_TOP_VALS[1:] - X_BL_TOP_VALS[:-1])**2 + (Y_BL_TOP[1:] - Y_BL_TOP[:-1])**2),axis = 0)
    first_idx       = np.ma.count_masked(X_BL_TOP,axis = 0)
    mask_count      = np.ma.count(X_BL_TOP,axis = 0)
    prev_index      = first_idx-1
    first_panel     = list(prev_index.flatten())
    last_panel      = list((first_idx-1 + mask_count).flatten())
    last_paneldve   = list((first_idx-2 + mask_count).flatten())
    aoas            = list(np.repeat(np.arange(ncases),ncpts))
    res             = list(np.tile(np.arange(ncpts),ncases) )
    X_BL_TOP.mask[first_panel,aoas,res] = False
    
    # flow velocity and pressure of on botton surface 
    VE_BL_TOP    = VT_BL
    CP_BL_TOP    = 1 - VE_BL_TOP**2     
    
    CP_BL_VALS   = np.ma.concatenate([np.flip(CP_BL_BOT,axis = 0),CP_BL_TOP], axis = 0 )  
    CP_BL_VALS_1 = CP_BL_VALS.flatten('F')  
    CP_BL_VALS_2 = CP_BL_VALS_1.data[~CP_BL_VALS_1.mask] 
    CP_BL        = CP_BL_VALS_2.reshape((npanel_mod,ncases,ncpts),order = 'F')    
    DCP_DX       = np.diff(CP_BL,axis=0)/ np.diff(X_BL,axis=0) 
 
    AERO_RES     = aero_coeff(x_coord_3d,y_coord_3d,CP,alpha,npanel) 
    
    # Squire-Young relation for total drag
    del2_inf_l   = THETA[0,:,:]*VE[0,:,:]**((5+H[0,:,:])/2)
    del2_inf_u   = THETA[-1,:,:]*VE[-1,:,:]**((5+H[-1,:,:])/2)
    del2_inf     = del2_inf_u + del2_inf_l
    cd_sqy        = 2*del2_inf.T
    
    # chordwise distribution of lift and drag
    fL, fD = chordwise_distribution(x_coord_3d,y_coord_3d,CP,alpha,npanel,CF,vt)
    
    airfoil_properties_old = Data(
        AoA            = alpha,
        Re             = Re_L,
        cl_invisc      = AERO_RES.cl,  
        cd_invisc      = AERO_RES.cdpi, 
        cm_invisc      = AERO_RES.cm,
        cd_visc        = cd_sqy,
        normals        = np.transpose(normals,(3,2,0,1)),
        x              = np.transpose(X,(2,1,0)),
        y              = np.transpose(Y,(2,1,0)),
        x_bl           = np.transpose(X_BL ,(2,1,0)),
        y_bl           = np.transpose(Y_BL ,(2,1,0)),
        cp             = np.transpose(CP_BL,(2,1,0)),  
        dcp_dx         = np.transpose(DCP_DX,(2,1,0)),            
        Ue_Vinf        = np.transpose(VE   ,(2,1,0)),         
        dVe            = np.transpose(DVE  ,(2,1,0)),   
        theta          = np.transpose(THETA,(2,1,0)),      
        delta_star     = np.transpose(DELTA_STAR,(2,1,0)),  
        delta          = np.transpose(DELTA,(2,1,0)),  
        Re_theta       = np.transpose(RE_THETA,(2,1,0)),  
        Re_x           = np.transpose(RE_X,(2,1,0)),  
        H              = np.transpose(H,(2,1,0)),            
        cf             = np.transpose(CF,(2,1,0)),
        fL             = fL,
        fD             = fD
        )  
        
    return  airfoil_properties_old


def concatenate_surfaces(X_BOT,X_TOP,FUNC_BOT_SURF,FUNC_TOP_SURF,npanel,ncases,ncpts): 
    '''Interpolation of airfoil properties   
    
    Assumptions:
    None

    Source:
    None                                                                    
                                                                   
    Inputs:                                    
    X_BOT          - bottom surface of airfoil                                     [unitless]
    X_TOP          - top surface of airfoil                                        [unitless]
    FUNC_BOT_SURF  - airfoil property computation discretization on bottom surface [multiple units]
    FUNC_TOP_SURF  - airfoil property computation discretization on top surface    [multiple units]
    npanel         - number of panels                                              [unitless]
    ncases         - number of angle of attacks                                    [unitless]
    ncpts          - number of Reynolds numbers                                    [unitless]
                                                                 
    Outputs:                                           
    FUNC           - airfoil property in user specified discretization on entire
                     surface of airfoil                                            [multiple units]
      
    Properties Used:
    N/A  
    '''  
    FUNC = np.zeros((npanel,ncases,ncpts))  
    
    for case in range(ncases):
        for cpt in range(ncpts):   
            top_func          = FUNC_TOP_SURF[:,case,cpt][X_TOP[:,case,cpt].mask == False] 
            bot_func          = FUNC_BOT_SURF[:,case,cpt][X_BOT[:,case,cpt].mask == False]                  
            FUNC[:,case,cpt]  = np.concatenate([bot_func[::-1],top_func])
    return FUNC




# ----------------------------------------------------------------------
#  Correlation-based methods for wing fuel capacity estimation
# ----------------------------------------------------------------------
## @ingroup Methods-Geometry-Two_Dimensional-Cross_Section-Planform
def wing_fuel_volume(wing):
    """Calculates the available fuel volume in a wing.

    Assumptions:
    None

    Source:
    Torenbeek, E., "Advanced Aircraft Design", 2013 (equation 10.30)

    Inputs:
    wing.
      areas.reference    [m^2]
      aspect_ratio       [-]
      thickness_to_chord [-]

    Outputs:
    wing.volume          [m^3]

    Properties Used:
    N/A
    """              

    # Unpack
    sref  = wing.areas.reference
    ar    = wing.aspect_ratio
    tc    = wing.thickness_to_chord

    # Calculate
    volume = 0.90* tc * sref** 1.5 * ar**-0.5 * 0.55

    # Pack
    wing.fuel_volume = volume